# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       Cloudflare
       cloudflare
       dns
"""
import logging


try:
    import httpx
except ImportError:     # pragma: no cover
    pass

from typing import Tuple, Optional

from ..http.proxy import HttpProxyBasePlugin
from ..common.flag import flags
from ..common.types import HostPort


logger = logging.getLogger(__name__)


flags.add_argument(
    '--cloudflare-dns-mode',
    type=str,
    default='security',
    help='Default: security.  Either "security" (for malware protection) ' +
    'or "family" (for malware and adult content protection)',
)


class CloudflareDnsResolverPlugin(HttpProxyBasePlugin):
    """This plugin uses Cloudflare DNS resolver to provide protection
    against malware and adult content.  Implementation uses :term:`DoH`
    specification.

    See https://developers.cloudflare.com/1.1.1.1/1.1.1.1-for-families
    See https://developers.cloudflare.com/1.1.1.1/encrypted-dns/dns-over-https/make-api-requests/dns-json

    .. note::

       For this plugin to work, make sure to bypass proxy for 1.1.1.1

    .. note::

       This plugin requires additional dependency because :term:`DoH`
       mandates a HTTP2 complaint client.  Install `httpx` dependency
       as::

           pip install "httpx[http2]"
    """

    def resolve_dns(self, host: str, port: int) -> Tuple[Optional[str], Optional[HostPort]]:
        try:
            context = httpx.create_ssl_context(http2=True)
            # TODO: Support resolution via Authority (SOA) to add support for
            # AAAA (IPv6) query
            r = httpx.get(
                'https://{0}.cloudflare-dns.com/dns-query?name={1}&type=A'.format(
                    self.flags.cloudflare_dns_mode, host,
                ),
                headers={'accept': 'application/dns-json'},
                verify=context,
                timeout=httpx.Timeout(timeout=5.0),
                proxies={
                    'all://': None,
                },
            )
            if r.status_code != 200:
                return None, None
            response = r.json()
            answers = response.get('Answer', [])
            if len(answers) == 0:
                return None, None
            # TODO: Utilize TTL to cache response locally
            # instead of making a DNS query repeatedly for the same host.
            return answers[0]['data'], None
        except Exception as e:
            logger.info(
                'Unable to resolve DNS-over-HTTPS for host {0} : {1}'.format(
                    host, str(e),
                ),
            )
            return None, None



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import time
import threading
from queue import Empty
from typing import Any, Deque
from collections import deque


class cached_property:      # pragma: no cover
    """Decorator for read-only properties evaluated only once within TTL period.
    It can be used to create a cached property like this::

        import random

        # the class containing the property must be a new-style class
        class MyClass:
            # create property whose value is cached for ten minutes
            @cached_property(ttl=600)
            def randint(self):
                # will only be evaluated every 10 min. at maximum.
                return random.randint(0, 100)

    The value is cached  in the '_cached_properties' attribute of the object instance that
    has the property getter method wrapped by this decorator. The '_cached_properties'
    attribute value is a dictionary which has a key for every property of the
    object which is wrapped by this decorator. Each entry in the cache is
    created only when the property is accessed for the first time and is a
    two-element tuple with the last computed property value and the last time
    it was updated in seconds since the epoch.

    The default time-to-live (TTL) is 0 seconds i.e. cached value will never expire.

    To expire a cached property value manually just do::
        del instance._cached_properties[<property name>]

    Adopted from https://wiki.python.org/moin/PythonDecoratorLibrary#Cached_Properties
    © 2011 Christopher Arndt, MIT License.

    NOTE: We need this function only because Python in-built are only available
    for 3.8+.  Hence, we must get rid of this function once proxy.py no longer
    support version older than 3.8.

    .. spelling::

       backports
       getter
       Arndt
       del
    """

    def __init__(self, ttl: float = 0):
        self.ttl = ttl

    def __call__(self, fget: Any, doc: Any = None) -> 'cached_property':
        self.fget = fget
        self.__doc__ = doc or fget.__doc__
        self.__name__ = fget.__name__
        self.__module__ = fget.__module__
        return self

    def __get__(self, inst: Any, owner: Any) -> Any:
        now = time.time()
        try:
            value, last_update = inst._cached_properties[self.__name__]
            if self.ttl > 0 and now - last_update > self.ttl:   # noqa: WPS333
                raise AttributeError
        except (KeyError, AttributeError):
            value = self.fget(inst)
            try:
                cache = inst._cached_properties
            except AttributeError:
                cache, inst._cached_properties = {}, {}
            finally:
                cache[self.__name__] = (value, now)  # pylint: disable=E0601
        return value


class NonBlockingQueue:
    '''Simple, unbounded, non-blocking FIFO queue.

    Supports only a single consumer.

    NOTE: This is available in Python since 3.7 as SimpleQueue.
    Here because proxy.py still supports 3.6
    '''

    def __init__(self) -> None:
        self._queue: Deque[Any] = deque()
        self._count: threading.Semaphore = threading.Semaphore(0)

    def put(self, item: Any) -> None:
        '''Put the item on the queue.'''
        self._queue.append(item)
        self._count.release()

    def get(self) -> Any:
        '''Remove and return an item from the queue.'''
        if not self._count.acquire(False, None):
            raise Empty
        return self._queue.popleft()

    def empty(self) -> bool:
        '''Return True if the queue is empty, False otherwise (not reliable!).'''
        return len(self._queue) == 0    # pragma: no cover

    def qsize(self) -> int:
        '''Return the approximate size of the queue (not reliable!).'''
        return len(self._queue)     # pragma: no cover



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import queue
import asyncio
import contextlib
from typing import Any, Optional

from .fd import ThreadlessFdExecutor
from ....common.backports import NonBlockingQueue


class LocalFdExecutor(ThreadlessFdExecutor[NonBlockingQueue]):
    """A threadless executor implementation which uses a queue to receive new work."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._loop: Optional[asyncio.AbstractEventLoop] = None

    @property
    def loop(self) -> Optional[asyncio.AbstractEventLoop]:
        if self._loop is None:
            self._loop = asyncio.get_event_loop_policy().new_event_loop()
        return self._loop

    def work_queue_fileno(self) -> Optional[int]:
        return None

    def receive_from_work_queue(self) -> bool:
        with contextlib.suppress(queue.Empty):
            work = self.work_queue.get()
            if isinstance(work, bool) and work is False:
                return True
            self.initialize(work)
        return False

    def initialize(self, work: Any) -> None:
        assert isinstance(work, tuple)
        conn, addr = work
        # NOTE: Here we are assuming to receive a connection object
        # and not a fileno because we are a LocalExecutor.
        fileno = conn.fileno()
        self.work(fileno, addr, conn)



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
from typing import Optional

from ..http.proxy import HttpProxyBasePlugin
from ..http.responses import okResponse


class ManInTheMiddlePlugin(HttpProxyBasePlugin):
    """Modifies upstream server responses."""

    def handle_upstream_chunk(self, _chunk: memoryview) -> Optional[memoryview]:
        return okResponse(content=b'Hello from man in the middle')



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import binascii


def pretty_hexlify(raw: bytes) -> str:
    hexlified = binascii.hexlify(raw).decode('utf-8')
    return ' '.join([hexlified[i: i+2] for i in range(0, len(hexlified), 2)])



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import time
import contextlib
from typing import Any, List, Optional, Generator

import unittest

from ..proxy import Proxy
from ..plugin import CacheResponsesPlugin
from ..common.utils import new_socket_connection
from ..common.constants import DEFAULT_TIMEOUT


class TestCase(unittest.TestCase):
    """Base TestCase class that automatically setup and tear down proxy.py."""

    DEFAULT_PROXY_PY_STARTUP_FLAGS = [
        '--hostname', '127.0.0.1',
        '--port', '0',
        '--num-workers', '1',
        '--num-acceptors', '1',
    ]

    PROXY: Optional[Proxy] = None
    INPUT_ARGS: Optional[List[str]] = None

    @classmethod
    def setUpClass(cls) -> None:
        cls.INPUT_ARGS = getattr(cls, 'PROXY_PY_STARTUP_FLAGS') \
            if hasattr(cls, 'PROXY_PY_STARTUP_FLAGS') \
            else cls.DEFAULT_PROXY_PY_STARTUP_FLAGS
        cls.PROXY = Proxy(cls.INPUT_ARGS)
        cls.PROXY.flags.plugins[b'HttpProxyBasePlugin'].append(
            CacheResponsesPlugin,
        )
        # pylint: disable=unnecessary-dunder-call
        cls.PROXY.__enter__()
        assert cls.PROXY.acceptors
        cls.wait_for_server(cls.PROXY.flags.port)

    @staticmethod
    def wait_for_server(
        proxy_port: int,
        wait_for_seconds: float = DEFAULT_TIMEOUT,
    ) -> None:
        """Wait for proxy.py server to come up."""
        start_time = time.time()
        while True:
            try:
                new_socket_connection(
                    ('localhost', proxy_port),
                ).close()
                break
            except ConnectionRefusedError:
                time.sleep(0.1)

            if time.time() - start_time > wait_for_seconds:
                raise TimeoutError(
                    'Timed out while waiting for proxy.py to start...',
                )

    @classmethod
    def tearDownClass(cls) -> None:
        assert cls.PROXY
        cls.PROXY.__exit__(None, None, None)
        cls.PROXY = None
        cls.INPUT_ARGS = None

    @contextlib.contextmanager
    def vcr(self) -> Generator[None, None, None]:
        try:
            CacheResponsesPlugin.ENABLED.set()
            yield
        finally:
            CacheResponsesPlugin.ENABLED.clear()

    def run(self, result: Optional[unittest.TestResult] = None) -> Any:
        super().run(result)



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       conn
"""
from typing import TYPE_CHECKING, Any

from .base import HttpProtocolException
from ..responses import BAD_GATEWAY_RESPONSE_PKT


if TYPE_CHECKING:   # pragma: no cover
    from ..parser import HttpParser


class ProxyConnectionFailed(HttpProtocolException):
    """Exception raised when ``HttpProxyPlugin`` is unable to establish connection to upstream server."""

    def __init__(self, host: str, port: int, reason: str, **kwargs: Any):
        self.host: str = host
        self.port: int = port
        self.reason: str = reason
        super().__init__('%s %s' % (self.__class__.__qualname__, reason), **kwargs)

    def response(self, _request: 'HttpParser') -> memoryview:
        return BAD_GATEWAY_RESPONSE_PKT



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import os
import asyncio
import logging
import argparse
import selectors
import multiprocessing
from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING, Any, Set, Dict, List, Tuple, Generic, TypeVar, Optional,
    cast,
)

from ...common.flag import flags as proxy_flags
from ...common.types import Readables, Writables, SelectableEvents
from ...common.logger import Logger
from ...common.constants import (
    DEFAULT_WAIT_FOR_TASKS_TIMEOUT, DEFAULT_SELECTOR_SELECT_TIMEOUT,
    DEFAULT_INACTIVE_CONN_CLEANUP_TIMEOUT,
)


if TYPE_CHECKING:   # pragma: no cover
    from .work import Work
    from ..event import EventQueue

T = TypeVar('T')

logger = logging.getLogger(__name__)


proxy_flags.add_argument(
    '--inactive-conn-cleanup-timeout',
    default=DEFAULT_INACTIVE_CONN_CLEANUP_TIMEOUT,
    help='Time after which inactive works must be cleaned up. '
    + 'Increase this value if your backend services are slow to response '
    + 'or when proxy.py is handling a high volume. When running proxy.py on Google Cloud (GCP) '
    + "you may see 'backend_connection_closed_before_data_sent_to_client', with curl clients "
    + "you may see 'Empty reply from server' error when '--inactive-conn-cleanup-timeout' "
    + 'value is low for your use-case. Default {0} seconds'.format(
        DEFAULT_INACTIVE_CONN_CLEANUP_TIMEOUT,
    ),
)


class Threadless(ABC, Generic[T]):
    """Work executor base class.

    Threadless provides an event loop, which is shared across
    multiple :class:`~proxy.core.acceptor.work.Work` instances to handle
    work.

    Threadless takes input a `work_klass` and an `event_queue`.  `work_klass`
    must conform to the :class:`~proxy.core.acceptor.work.Work`
    protocol.  Work is received over the `event_queue`.

    When a work is accepted, threadless creates a new instance of `work_klass`.
    Threadless will then invoke necessary lifecycle of the
    :class:`~proxy.core.acceptor.work.Work` protocol,
    allowing `work_klass` implementation to handle the assigned work.

    Example, :class:`~proxy.core.base.tcp_server.BaseTcpServerHandler`
    implements :class:`~proxy.core.acceptor.work.Work` protocol. It
    expects a client connection as work payload and hooks into the
    threadless event loop to handle the client connection.
    """

    def __init__(
            self,
            iid: str,
            work_queue: T,
            flags: argparse.Namespace,
            event_queue: Optional['EventQueue'] = None,
    ) -> None:
        super().__init__()
        self.iid = iid
        self.work_queue = work_queue
        self.flags = flags
        self.event_queue = event_queue

        self.running = multiprocessing.Event()
        self.works: Dict[int, 'Work[Any]'] = {}
        self.selector: Optional[selectors.DefaultSelector] = None
        # If we remove single quotes for typing hint below,
        # runtime exceptions will occur for < Python 3.9.
        #
        # Ref https://github.com/abhinavsingh/proxy.py/runs/4279055360?check_suite_focus=true
        self.unfinished: Set['asyncio.Task[bool]'] = set()
        self.registered_events_by_work_ids: Dict[
            # work_id
            int,
            # fileno, mask
            SelectableEvents,
        ] = {}
        self.wait_timeout: float = DEFAULT_WAIT_FOR_TASKS_TIMEOUT
        self.cleanup_inactive_timeout: float = float(
            self.flags.inactive_conn_cleanup_timeout,
        )
        self._total: int = 0
        # When put at the top, causes circular import error
        # since integrated ssh tunnel was introduced.
        from ..connection import (  # pylint: disable=C0415
            UpstreamConnectionPool,
        )
        self._upstream_conn_pool: Optional['UpstreamConnectionPool'] = None
        self._upstream_conn_filenos: Set[int] = set()
        if self.flags.enable_conn_pool:
            self._upstream_conn_pool = UpstreamConnectionPool()

    @property
    @abstractmethod
    def loop(self) -> Optional[asyncio.AbstractEventLoop]:
        raise NotImplementedError()

    @abstractmethod
    def receive_from_work_queue(self) -> bool:
        """Work queue is ready to receive new work.

        Receive it and call ``work_on_tcp_conn``.

        Return True to tear down the loop."""
        raise NotImplementedError()

    @abstractmethod
    def work_queue_fileno(self) -> Optional[int]:
        """If work queue must be selected before calling
        ``receive_from_work_queue`` then implementation must
        return work queue fd."""
        raise NotImplementedError()

    @abstractmethod
    def work(self, *args: Any) -> None:
        raise NotImplementedError()

    def create(self, uid: str, *args: Any) -> 'Work[T]':
        return cast(
            'Work[T]', self.flags.work_klass(
                self.flags.work_klass.create(*args),
                flags=self.flags,
                event_queue=self.event_queue,
                uid=uid,
                upstream_conn_pool=self._upstream_conn_pool,
            ),
        )

    def close_work_queue(self) -> None:
        """Only called if ``work_queue_fileno`` returns an integer.
        If an fd is select-able for work queue, make sure
        to close the work queue fd now."""
        pass    # pragma: no cover

    async def _update_work_events(self, work_id: int) -> None:
        assert self.selector is not None
        worker_events = await self.works[work_id].get_events()
        # NOTE: Current assumption is that multiple works will not
        # be interested in the same fd.  Descriptors of interests
        # returned by work must be unique.
        #
        # TODO: Ideally we must diff and unregister socks not
        # returned of interest within current _select_events call
        # but exists in the registered_socks_by_work_ids registry.
        for fileno in worker_events:
            if work_id not in self.registered_events_by_work_ids:
                self.registered_events_by_work_ids[work_id] = {}
            mask = worker_events[fileno]
            if fileno in self.registered_events_by_work_ids[work_id]:
                oldmask = self.registered_events_by_work_ids[work_id][fileno]
                if mask != oldmask:
                    self.selector.modify(
                        fileno, events=mask,
                        data=work_id,
                    )
                    self.registered_events_by_work_ids[work_id][fileno] = mask
                    logger.debug(
                        'fd#{0} modified for mask#{1} by work#{2}'.format(
                            fileno, mask, work_id,
                        ),
                    )
                # else:
                #     logger.info(
                #         'fd#{0} by work#{1} not modified'.format(fileno, work_id))
            elif fileno in self._upstream_conn_filenos:
                # Descriptor offered by work, but is already registered by connection pool
                # Most likely because work has acquired a reusable connection.
                self.selector.modify(fileno, events=mask, data=work_id)
                self.registered_events_by_work_ids[work_id][fileno] = mask
                self._upstream_conn_filenos.remove(fileno)
                logger.debug(
                    'fd#{0} borrowed with mask#{1} by work#{2}'.format(
                        fileno, mask, work_id,
                    ),
                )
            # Can throw ValueError: Invalid file descriptor: -1
            #
            # A guard within Work classes may not help here due to
            # asynchronous nature.  Hence, threadless will handle
            # ValueError exceptions raised by selector.register
            # for invalid fd.
            #
            # TODO: Also remove offending work from pool to avoid spin loop.
            elif fileno != -1:
                try:
                    self.selector.register(fileno, events=mask, data=work_id)
                    self.registered_events_by_work_ids[work_id][fileno] = mask
                    logger.debug(
                        'fd#{0} registered for mask#{1} by work#{2}'.format(
                            fileno,
                            mask,
                            work_id,
                        ),
                    )
                except KeyError as exc:
                    logger.warning(
                        'KeyError when trying to register fd#{0}'.format(fileno),
                        exc_info=exc,
                    )

    async def _update_conn_pool_events(self) -> None:
        if not self._upstream_conn_pool:
            return
        assert self.selector is not None
        new_conn_pool_events = await self._upstream_conn_pool.get_events()
        old_conn_pool_filenos = self._upstream_conn_filenos.copy()
        self._upstream_conn_filenos.clear()
        new_conn_pool_filenos = set(new_conn_pool_events.keys())
        new_conn_pool_filenos.difference_update(old_conn_pool_filenos)
        for fileno in new_conn_pool_filenos:
            self.selector.register(
                fileno,
                events=new_conn_pool_events[fileno],
                data=0,
            )
            self._upstream_conn_filenos.add(fileno)
        old_conn_pool_filenos.difference_update(self._upstream_conn_filenos)
        for fileno in old_conn_pool_filenos:
            self.selector.unregister(fileno)

    async def _update_selector(self) -> None:
        assert self.selector is not None
        unfinished_work_ids = set()
        for task in self.unfinished:
            unfinished_work_ids.add(task._work_id)   # type: ignore
        for work_id in self.works:
            # We don't want to invoke work objects which haven't
            # yet finished their previous task
            if work_id in unfinished_work_ids:
                continue
            await self._update_work_events(work_id)
        await self._update_conn_pool_events()

    async def _selected_events(self) -> Tuple[
            Dict[int, Tuple[Readables, Writables]],
            bool,
    ]:
        """For each work, collects events that they are interested in.
        Calls select for events of interest.

        Returns a 2-tuple containing a dictionary and boolean.
        Dictionary keys are work IDs and values are 2-tuple
        containing ready readables & writables.

        Returned boolean value indicates whether there is
        a newly accepted work waiting to be received and
        queued for processing.  This is only applicable when
        :class:`~proxy.core.work.threadless.Threadless.work_queue_fileno`
        returns a valid fd.
        """
        assert self.selector is not None
        await self._update_selector()
        # Keys are work_id and values are 2-tuple indicating
        # readables & writables that work_id is interested in
        # and are ready for IO.
        work_by_ids: Dict[int, Tuple[Readables, Writables]] = {}
        new_work_available = False
        wqfileno = self.work_queue_fileno()
        if wqfileno is None:
            # When ``work_queue_fileno`` returns None,
            # always return True for the boolean value.
            new_work_available = True

        events = self.selector.select(
            timeout=DEFAULT_SELECTOR_SELECT_TIMEOUT,
        )

        for key, mask in events:
            if not new_work_available and wqfileno is not None and key.fileobj == wqfileno:
                assert mask & selectors.EVENT_READ
                new_work_available = True
                continue
            if key.data not in work_by_ids:
                work_by_ids[key.data] = ([], [])
            if mask & selectors.EVENT_READ:
                work_by_ids[key.data][0].append(key.fd)
            if mask & selectors.EVENT_WRITE:
                work_by_ids[key.data][1].append(key.fd)
        return (work_by_ids, new_work_available)

    async def _wait_for_tasks(self) -> Set['asyncio.Task[bool]']:
        finished, self.unfinished = await asyncio.wait(
            self.unfinished,
            timeout=self.wait_timeout,
            return_when=asyncio.FIRST_COMPLETED,
        )
        return finished     # noqa: WPS331

    def _cleanup_inactive(self) -> None:
        inactive_works: List[int] = []
        for work_id in self.works:
            if self.works[work_id].is_inactive():
                inactive_works.append(work_id)
        for work_id in inactive_works:
            self._cleanup(work_id)

    # TODO: HttpProtocolHandler.shutdown can call flush which may block
    def _cleanup(self, work_id: int) -> None:
        if work_id in self.registered_events_by_work_ids:
            assert self.selector
            for fileno in self.registered_events_by_work_ids[work_id]:
                logger.debug(
                    'fd#{0} unregistered by work#{1}'.format(
                        fileno, work_id,
                    ),
                )
                self.selector.unregister(fileno)
            self.registered_events_by_work_ids[work_id].clear()
            del self.registered_events_by_work_ids[work_id]
        try:
            self.works[work_id].shutdown()
        except Exception as exc:
            logger.exception(
                'Error when shutting down work#{0}'.format(work_id),
                exc_info=exc,
            )
        finally:
            del self.works[work_id]
            if self.work_queue_fileno() is not None:
                os.close(work_id)

    def _create_tasks(
            self,
            work_by_ids: Dict[int, Tuple[Readables, Writables]],
    ) -> Set['asyncio.Task[bool]']:
        assert self.loop
        tasks: Set['asyncio.Task[bool]'] = set()
        for work_id in work_by_ids:
            if work_id == 0:
                assert self._upstream_conn_pool
                task = self.loop.create_task(
                    self._upstream_conn_pool.handle_events(
                        *work_by_ids[work_id],
                    ),
                )
            else:
                task = self.loop.create_task(
                    self.works[work_id].handle_events(*work_by_ids[work_id]),
                )
            task._work_id = work_id     # type: ignore[attr-defined]
            # task.set_name(work_id)
            tasks.add(task)
        return tasks

    async def _run_once(self) -> bool:
        assert self.loop is not None
        work_by_ids, new_work_available = await self._selected_events()
        # Accept new work if available
        #
        # TODO: We must use a work klass to handle
        # client_queue fd itself a.k.a. accept_client
        # will become handle_readables.
        if new_work_available:
            teardown = self.receive_from_work_queue()
            if teardown:
                return teardown
        if len(work_by_ids) == 0:
            return False
        # Invoke Threadless.handle_events
        self.unfinished.update(self._create_tasks(work_by_ids))
        # logger.debug('Executing {0} works'.format(len(self.unfinished)))
        # Cleanup finished tasks
        for task in await self._wait_for_tasks():
            # Checking for result can raise exception e.g.
            # CancelledError, InvalidStateError or an exception
            # from underlying task e.g. TimeoutError.
            teardown = False
            work_id = task._work_id     # type: ignore
            try:
                teardown = task.result()
            except Exception:
                teardown = True
            finally:
                if teardown:
                    self._cleanup(work_id)
                    # self.cleanup(int(task.get_name()))
        # logger.debug(
        #     'Done executing works, {0} pending, {1} registered'.format(
        #         len(self.unfinished), len(self.registered_events_by_work_ids),
        #     ),
        # )
        return False

    async def _run_forever(self) -> None:
        tick = 0
        try:
            while True:
                if await self._run_once():
                    break
                # Check for inactive and shutdown signal
                elapsed = tick * \
                    (DEFAULT_SELECTOR_SELECT_TIMEOUT + self.wait_timeout)
                if elapsed >= self.cleanup_inactive_timeout:
                    self._cleanup_inactive()
                    if self.running.is_set():
                        break
                    tick = 0
                tick += 1
        except KeyboardInterrupt:
            pass
        finally:
            if self.loop:
                self.loop.stop()

    def run(self) -> None:
        Logger.setup(
            self.flags.log_file, self.flags.log_level,
            self.flags.log_format,
        )
        wqfileno = self.work_queue_fileno()
        try:
            self.selector = selectors.DefaultSelector()
            if wqfileno is not None:
                self.selector.register(
                    wqfileno,
                    selectors.EVENT_READ,
                    data=wqfileno,
                )
            assert self.loop
            logger.debug('Working on {0} works'.format(len(self.works)))
            self.loop.create_task(self._run_forever())
            self.loop.run_forever()
        except KeyboardInterrupt:
            pass
        finally:
            assert self.selector is not None
            if wqfileno is not None:
                self.selector.unregister(wqfileno)
                self.close_work_queue()
            self.selector.close()
            assert self.loop is not None
            self.loop.run_until_complete(self.loop.shutdown_asyncgens())
            self.loop.close()



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
from typing import TYPE_CHECKING, Optional
from multiprocessing.reduction import send_handle


if TYPE_CHECKING:   # pragma: no cover
    import socket
    import multiprocessing
    from multiprocessing import connection

    from ...common.types import HostPort


def delegate_work_to_pool(
        worker_pid: int,
        work_queue: 'connection.Connection',
        work_lock: 'multiprocessing.synchronize.Lock',
        conn: 'socket.socket',
        addr: Optional['HostPort'],
        unix_socket_path: Optional[str] = None,
) -> None:
    """Utility method to delegate a work to threadless executor pool."""
    with work_lock:
        # Accepted client address is empty string for
        # unix socket domain, avoid sending empty string
        # for optimization.
        if not unix_socket_path:
            work_queue.send(addr)
        send_handle(
            work_queue,
            conn.fileno(),
            worker_pid,
        )
        conn.close()



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import socket
import argparse
import threading
from typing import TYPE_CHECKING, Tuple, TypeVar, Optional

from ..event import EventQueue, eventNames


if TYPE_CHECKING:   # pragma: no cover
    from .work import Work
    from ...common.types import HostPort

T = TypeVar('T')


# TODO: Add generic T
def start_threaded_work(
        flags: argparse.Namespace,
        conn: socket.socket,
        addr: Optional['HostPort'],
        event_queue: Optional[EventQueue] = None,
        publisher_id: Optional[str] = None,
) -> Tuple['Work[T]', threading.Thread]:
    """Utility method to start a work in a new thread."""
    work = flags.work_klass(
        flags.work_klass.create(conn, addr),
        flags=flags,
        event_queue=event_queue,
        upstream_conn_pool=None,
    )
    # TODO: Keep reference to threads and join during shutdown.
    # This will ensure connections are not abruptly closed on shutdown
    # for threaded execution mode.
    thread = threading.Thread(target=work.run)
    thread.daemon = True
    thread.start()
    work.publish_event(
        event_name=eventNames.WORK_STARTED,
        event_payload={'fileno': conn.fileno(), 'addr': addr},
        publisher_id=publisher_id or 'thread#{0}'.format(
            thread.ident,
        ),
    )
    return (work, thread)



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
from .web import HttpWebServerPlugin
from .plugin import ReverseProxyBasePlugin, HttpWebServerBasePlugin
from .metrics import MetricsWebServerPlugin
from .protocols import httpProtocolTypes
from .pac_plugin import HttpWebServerPacFilePlugin


__all__ = [
    'HttpWebServerPlugin',
    'HttpWebServerPacFilePlugin',
    'HttpWebServerBasePlugin',
    'httpProtocolTypes',
    'ReverseProxyBasePlugin',
    'MetricsWebServerPlugin',
]



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import os
import subprocess
from typing import Any, Dict, Optional

from ..http.proxy import HttpProxyBasePlugin
from ..http.parser import HttpParser
from ..common.utils import text_
from ..common.constants import IS_WINDOWS


class ProgramNamePlugin(HttpProxyBasePlugin):
    """Tries to identify the application connecting to the
    proxy instance.  This is only possible when connection
    itself originates from the same machine where the proxy
    instance is running."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.program_name: Optional[str] = None

    def before_upstream_connection(
            self, request: HttpParser,
    ) -> Optional[HttpParser]:
        if IS_WINDOWS:
            raise NotImplementedError()
        assert self.client.addr
        if self.client.addr[0] in ('::1', '127.0.0.1'):
            port = self.client.addr[1]
            ls = subprocess.Popen(
                ('lsof', '-i', '-P', '-n'),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            try:
                output = subprocess.check_output(
                    ('grep', '{0}'.format(port)),
                    stdin=ls.stdout,
                )
                port_programs = output.splitlines()
                for program in port_programs:
                    parts = program.split()
                    if int(parts[1]) != os.getpid():
                        self.program_name = text_(parts[0])
                        break
            except subprocess.CalledProcessError:
                pass
            finally:
                ls.wait(timeout=1)
        if self.program_name is None:
            self.program_name = self.client.addr[0]
        return request

    def on_access_log(self, context: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        context.update({'client_ip': self.program_name})
        return context



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       url
"""
import re
import json
import logging
from typing import Any, Dict, List, Optional

from ..http import httpStatusCodes
from ..http.proxy import HttpProxyBasePlugin
from ..common.flag import flags
from ..http.parser import HttpParser
from ..common.utils import text_
from ..http.exception import HttpRequestRejected


logger = logging.getLogger(__name__)

# See adblock.json file in repository for sample example config
flags.add_argument(
    '--filtered-url-regex-config',
    type=str,
    default='',
    help='Default: No config.  Comma separated list of IPv4 and IPv6 addresses.',
)


class FilterByURLRegexPlugin(HttpProxyBasePlugin):
    """Drops traffic by inspecting request URL and checking
    against a list of regular expressions.  Example, default
    filter list below can be used as a starting point for
    filtering ads.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.filters: List[Dict[str, Any]] = []
        if self.flags.filtered_url_regex_config != '':
            with open(self.flags.filtered_url_regex_config, 'rb') as f:
                self.filters = json.load(f)

    def handle_client_request(
            self, request: HttpParser,
    ) -> Optional[HttpParser]:
        # determine host
        request_host = None
        if request.host:
            request_host = request.host
        elif request.headers and b'host' in request.headers:
            request_host = request.header(b'host')

        if not request_host:
            logger.error("Cannot determine host")
            return request

        # build URL
        url = b'%s%s' % (
            request_host,
            request.path,
        )
        # check URL against list
        for rule_number, blocked_entry in enumerate(self.filters, start=1):
            # if regex matches on URL
            if re.search(text_(blocked_entry['regex']), text_(url)):
                # log that the request has been filtered
                logger.info(
                    "Blocked: %r with status_code '%r' by rule number '%r'" % (
                        text_(url),
                        httpStatusCodes.NOT_FOUND,
                        rule_number,
                    ),
                )
                # close the connection with the status code from the filter
                # list
                raise HttpRequestRejected(
                    status_code=httpStatusCodes.NOT_FOUND,
                    reason=b'Blocked',
                )
        return request



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       pki
"""
import os
import sys
import time
import uuid
import logging
import argparse
import tempfile
import contextlib
import subprocess
from typing import List, Tuple, Optional, Generator

from .utils import bytes_
from .version import __version__
from .constants import COMMA


logger = logging.getLogger(__name__)


DEFAULT_CONFIG = b'''[ req ]
#default_bits		= 2048
#default_md		    = sha256
#default_keyfile 	= privkey.pem
distinguished_name	= req_distinguished_name
attributes		    = req_attributes

[ req_distinguished_name ]
countryName			    = Country Name (2 letter code)
countryName_min			= 2
countryName_max			= 2
stateOrProvinceName		= State or Province Name (full name)
localityName			= Locality Name (eg, city)
organizationName		= Organization Name (eg, company)
organizationalUnitName	= Organizational Unit Name (eg, section)
commonName			    = Common Name (eg, fully qualified host name)
commonName_max			= 64
emailAddress			= Email Address
emailAddress_max		= 64

[ req_attributes ]
challengePassword		= A challenge password
challengePassword_min	= 4
challengePassword_max	= 20'''


def remove_passphrase(
        key_in_path: str,
        password: str,
        key_out_path: str,
        timeout: int = 10,
        openssl: str = 'openssl',
) -> bool:
    """Remove passphrase from a private key."""
    command = [
        openssl, 'rsa',
        '-passin', 'pass:%s' % password,
        '-in', key_in_path,
        '-out', key_out_path,
    ]
    return run_openssl_command(command, timeout)


def gen_private_key(
        key_path: str,
        password: str,
        bits: int = 2048,
        timeout: int = 10,
        openssl: str = 'openssl',
) -> bool:
    """Generates a private key."""
    command = [
        openssl, 'genrsa', '-aes256',
        '-passout', 'pass:%s' % password,
        '-out', key_path, str(bits),
    ]
    return run_openssl_command(command, timeout)


def gen_public_key(
        public_key_path: str,
        private_key_path: str,
        private_key_password: str,
        subject: str,
        alt_subj_names: Optional[List[str]] = None,
        extended_key_usage: Optional[str] = None,
        validity_in_days: int = 365,
        timeout: int = 10,
        openssl: str = 'openssl',
) -> bool:
    """For a given private key, generates a corresponding public key."""
    with ssl_config(alt_subj_names, extended_key_usage) as (config_path, has_extension):
        command = [
            openssl, 'req', '-new', '-x509', '-sha256',
            '-days', str(validity_in_days), '-subj', subject,
            '-passin', 'pass:%s' % private_key_password,
            '-config', config_path,
            '-key', private_key_path, '-out', public_key_path,
        ]
        if has_extension:
            command.extend([
                '-extensions', 'PROXY',
            ])
        return run_openssl_command(command, timeout)


def gen_csr(
        csr_path: str,
        key_path: str,
        password: str,
        crt_path: str,
        timeout: int = 10,
        openssl: str = 'openssl',
) -> bool:
    """Generates a CSR based upon existing certificate and key file."""
    command = [
        openssl, 'x509', '-x509toreq',
        '-passin', 'pass:%s' % password,
        '-in', crt_path, '-signkey', key_path,
        '-out', csr_path,
    ]
    return run_openssl_command(command, timeout)


def sign_csr(
        csr_path: str,
        crt_path: str,
        ca_key_path: str,
        ca_key_password: str,
        ca_crt_path: str,
        serial: str,
        alt_subj_names: Optional[List[str]] = None,
        extended_key_usage: Optional[str] = None,
        validity_in_days: int = 365,
        timeout: int = 10,
        openssl: str = 'openssl',
) -> bool:
    """Sign a CSR using CA key and certificate."""
    with ext_file(alt_subj_names, extended_key_usage) as extension_path:
        command = [
            openssl, 'x509', '-req', '-sha256',
            '-CA', ca_crt_path,
            '-CAkey', ca_key_path,
            '-passin', 'pass:%s' % ca_key_password,
            '-set_serial', serial,
            '-days', str(validity_in_days),
            '-extfile', extension_path,
            '-in', csr_path,
            '-out', crt_path,
        ]
        return run_openssl_command(command, timeout)


def get_ext_config(
        alt_subj_names: Optional[List[str]] = None,
        extended_key_usage: Optional[str] = None,
) -> bytes:
    config = b''
    # Add SAN extension
    if alt_subj_names is not None and len(alt_subj_names) > 0:
        alt_names = []
        for cname in alt_subj_names:
            alt_names.append(b'DNS:%s' % bytes_(cname))
        config += b'\nsubjectAltName=' + COMMA.join(alt_names)
    # Add extendedKeyUsage section
    if extended_key_usage is not None:
        config += b'\nextendedKeyUsage=' + bytes_(extended_key_usage)
    return config


@contextlib.contextmanager
def ext_file(
        alt_subj_names: Optional[List[str]] = None,
        extended_key_usage: Optional[str] = None,
) -> Generator[str, None, None]:
    # Write config to temp file
    config_path = os.path.join(tempfile.gettempdir(), uuid.uuid4().hex)
    with open(config_path, 'wb') as cnf:
        cnf.write(
            get_ext_config(alt_subj_names, extended_key_usage),
        )

    yield config_path

    # Delete temp file
    os.remove(config_path)


@contextlib.contextmanager
def ssl_config(
        alt_subj_names: Optional[List[str]] = None,
        extended_key_usage: Optional[str] = None,
) -> Generator[Tuple[str, bool], None, None]:
    config = DEFAULT_CONFIG

    has_extension = False
    if (alt_subj_names is not None and len(alt_subj_names) > 0) or \
            extended_key_usage is not None:
        has_extension = True
        config += b'\n[PROXY]'

    # Add custom extensions
    config += get_ext_config(alt_subj_names, extended_key_usage)

    # Write config to temp file
    config_path = os.path.join(tempfile.gettempdir(), uuid.uuid4().hex)
    with open(config_path, 'wb') as cnf:
        cnf.write(config)

    yield config_path, has_extension

    # Delete temp file
    os.remove(config_path)


def run_openssl_command(command: List[str], timeout: int) -> bool:
    cmd = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    cmd.communicate(timeout=timeout)
    return cmd.returncode == 0


if __name__ == '__main__':
    available_actions = (
        'remove_passphrase', 'gen_private_key', 'gen_public_key',
        'gen_csr', 'sign_csr',
    )

    parser = argparse.ArgumentParser(
        description='proxy.py v%s : PKI Utility' % __version__,
    )
    parser.add_argument(
        'action',
        type=str,
        default=None,
        help='Valid actions: ' + ', '.join(available_actions),
    )
    parser.add_argument(
        '--password',
        type=str,
        default='proxy.py',
        help='Password to use for encryption. Default: proxy.py',
    )
    parser.add_argument(
        '--private-key-path',
        type=str,
        default=None,
        help='Private key path',
    )
    parser.add_argument(
        '--public-key-path',
        type=str,
        default=None,
        help='Public key path',
    )
    parser.add_argument(
        '--subject',
        type=str,
        default='/CN=localhost',
        help='Subject to use for public key generation. Default: /CN=localhost',
    )
    parser.add_argument(
        '--csr-path',
        type=str,
        default=None,
        help='CSR file path.  Use with gen_csr and sign_csr action.',
    )
    parser.add_argument(
        '--crt-path',
        type=str,
        default=None,
        help='Signed certificate path.  Use with sign_csr action.',
    )
    parser.add_argument(
        '--hostname',
        type=str,
        default=None,
        help='Alternative subject names to use during CSR signing.',
    )
    parser.add_argument(
        '--openssl',
        type=str,
        default='openssl',
        help='Path to openssl binary.  By default, we assume openssl is in your PATH',
    )
    args = parser.parse_args(sys.argv[1:])

    # Validation
    if args.action not in available_actions:
        logger.error(
            'Invalid --action. Valid values ' +
            ', '.join(available_actions),
        )
        sys.exit(1)
    if args.action in ('gen_private_key', 'gen_public_key') and \
            args.private_key_path is None:
        logger.error('--private-key-path is required for ' + args.action)
        sys.exit(1)
    if args.action == 'gen_public_key' and \
            args.public_key_path is None:
        logger.error(
            '--public-key-file is required for private key generation',
        )
        sys.exit(1)

    # Execute
    if args.action == 'gen_private_key':
        gen_private_key(
            args.private_key_path,
            args.password, openssl=args.openssl,
        )
    elif args.action == 'gen_public_key':
        gen_public_key(
            args.public_key_path, args.private_key_path,
            args.password, args.subject, openssl=args.openssl,
        )
    elif args.action == 'remove_passphrase':
        remove_passphrase(
            args.private_key_path, args.password,
            args.private_key_path, openssl=args.openssl,
        )
    elif args.action == 'gen_csr':
        gen_csr(
            args.csr_path,
            args.private_key_path,
            args.password,
            args.public_key_path,
            openssl=args.openssl,
        )
    elif args.action == 'sign_csr':
        sign_csr(
            args.csr_path, args.crt_path, args.private_key_path, args.password,
            args.public_key_path, str(int(time.time())), alt_subj_names=[args.hostname],
            openssl=args.openssl,
        )



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
from typing import TYPE_CHECKING, Any, Dict, Optional

from .base import HttpProtocolException
from ...common.utils import build_http_response


if TYPE_CHECKING:   # pragma: no cover
    from ..parser import HttpParser


class HttpRequestRejected(HttpProtocolException):
    """Generic exception that can be used to reject the client requests.

    Connections can either be dropped/closed or optionally an
    HTTP status code can be returned."""

    def __init__(
            self,
            status_code: Optional[int] = None,
            reason: Optional[bytes] = None,
            headers: Optional[Dict[bytes, bytes]] = None,
            body: Optional[bytes] = None,
            **kwargs: Any,
    ):
        self.status_code: Optional[int] = status_code
        self.reason: Optional[bytes] = reason
        self.headers: Optional[Dict[bytes, bytes]] = headers
        self.body: Optional[bytes] = body
        klass_name = self.__class__.__qualname__
        super().__init__(
            message='%s %r' % (klass_name, reason)
            if reason
            else klass_name,
            **kwargs,
        )

    def response(self, _request: 'HttpParser') -> Optional[memoryview]:
        if self.status_code:
            return memoryview(
                build_http_response(
                    status_code=self.status_code,
                    reason=self.reason,
                    headers=self.headers,
                    body=self.body,
                    conn_close=True,
                ),
            )
        return None



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
from typing import Any, Optional

from ..http.proxy import HttpProxyBasePlugin
from ..http.parser import HttpParser, httpParserTypes


class ModifyChunkResponsePlugin(HttpProxyBasePlugin):
    """Accumulate & modify chunk responses as received from upstream."""

    DEFAULT_CHUNKS = [
        b'modify',
        b'chunk',
        b'response',
        b'plugin',
    ]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        # Create a new http protocol parser for response payloads
        self.response = HttpParser(httpParserTypes.RESPONSE_PARSER)

    def handle_upstream_chunk(self, chunk: memoryview) -> Optional[memoryview]:
        # Parse the response.
        # Note that these chunks also include headers
        self.response.parse(chunk)
        # If response is complete, modify and dispatch to client
        if self.response.is_complete:
            # Queue our custom chunk if response is chunked encoded
            # otherwise queue the original response to client
            if self.response.is_chunked_encoded:
                self.response.body = b'\n'.join(self.DEFAULT_CHUNKS) + b'\n'
            self.client.queue(memoryview(self.response.build_response()))
        # Avoid returning chunk straight to client
        return None



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import json
import argparse
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, List

from . import WebsocketFrame
from ...core.event import EventQueue
from ...common.utils import bytes_


if TYPE_CHECKING:   # pragma: no cover
    from ..connection import HttpClientConnection


class WebSocketTransportBasePlugin(ABC):
    """Abstract class for plugins extending dashboard websocket API."""

    def __init__(
            self,
            flags: argparse.Namespace,
            client: 'HttpClientConnection',
            event_queue: EventQueue,
    ) -> None:
        self.flags = flags
        self.client = client
        self.event_queue = event_queue

    @abstractmethod
    def methods(self) -> List[str]:
        """Return list of methods that this plugin will handle."""
        pass    # pragma: no cover

    def connected(self) -> None:
        """Invoked when client websocket handshake finishes."""
        pass    # pragma: no cover

    @abstractmethod
    def handle_message(self, message: Dict[str, Any]) -> None:
        """Handle messages for registered methods."""
        pass    # pragma: no cover

    def disconnected(self) -> None:
        """Invoked when client websocket connection gets closed."""
        pass    # pragma: no cover

    def reply(self, data: Dict[str, Any]) -> None:
        self.client.queue(
            memoryview(
                WebsocketFrame.text(
                    bytes_(
                        json.dumps(data),
                    ),
                ),
            ),
        )



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import queue
import logging
import threading
from typing import Any, Dict, List
from multiprocessing import connection

from .names import eventNames
from .queue import EventQueue


logger = logging.getLogger(__name__)


class EventDispatcher:
    """Core EventDispatcher.

    Direct consuming from global events queue outside of dispatcher
    module is not-recommended.  Python native multiprocessing queue
    doesn't provide a fanout functionality which core dispatcher module
    implements so that several plugins can consume the same published
    event concurrently (when necessary).

    When --enable-events is used, a multiprocessing.Queue is created and
    attached to global flags.  This queue can then be used for
    dispatching an Event dict object into the queue.

    When --enable-events is used, dispatcher module is automatically
    started.  Most importantly, dispatcher module ensures that queue is
    not flooded and doesn't utilize too much memory in case there are no
    event subscribers for published messages.

    EventDispatcher ensures that subscribers will receive the messages
    in the order they are published.
    """

    def __init__(
            self,
            shutdown: threading.Event,
            event_queue: EventQueue,
    ) -> None:
        self.shutdown: threading.Event = shutdown
        self.event_queue: EventQueue = event_queue
        # subscriber connection objects
        self.subscribers: Dict[str, connection.Connection] = {}

    def handle_event(self, ev: Dict[str, Any]) -> None:
        if ev['event_name'] == eventNames.SUBSCRIBE:
            sub_id = ev['event_payload']['sub_id']
            self.subscribers[sub_id] = ev['event_payload']['conn']
            # send ack
            if not self._send(
                sub_id, {
                    'event_name': eventNames.SUBSCRIBED,
                },
            ):
                self._close_and_delete(sub_id)
        elif ev['event_name'] == eventNames.UNSUBSCRIBE:
            sub_id = ev['event_payload']['sub_id']
            if sub_id in self.subscribers:
                # send ack
                logger.debug('unsubscription request ack sent')
                self._send(
                    sub_id, {
                        'event_name': eventNames.UNSUBSCRIBED,
                    },
                )
                self._close_and_delete(sub_id)
            else:
                logger.info(
                    'unsubscription request ack not sent, subscriber already gone',
                )
        else:
            # logger.info(ev)
            self._broadcast(ev)

    def run_once(self) -> None:
        ev: Dict[str, Any] = self.event_queue.queue.get(timeout=1)
        self.handle_event(ev)

    def run(self) -> None:
        try:
            while not self.shutdown.is_set():
                try:
                    self.run_once()
                except queue.Empty:
                    pass
        except KeyboardInterrupt:
            pass
        except Exception as e:
            logger.exception('Dispatcher exception', exc_info=e)
        finally:
            # Send shutdown message to all active subscribers
            self._broadcast({
                'event_name': eventNames.DISPATCHER_SHUTDOWN,
            })
            logger.info('Dispatcher shutdown')

    def _broadcast(self, ev: Dict[str, Any]) -> None:
        broken_pipes: List[str] = []
        for sub_id in self.subscribers:
            try:
                self.subscribers[sub_id].send(ev)
            except BrokenPipeError:
                logger.warning(
                    'Subscriber#%s broken pipe', sub_id,
                )
                self._close(sub_id)
                broken_pipes.append(sub_id)
        for sub_id in broken_pipes:
            del self.subscribers[sub_id]

    def _close_and_delete(self, sub_id: str) -> None:
        self._close(sub_id)
        del self.subscribers[sub_id]

    def _close(self, sub_id: str) -> None:
        try:
            self.subscribers[sub_id].close()
        except Exception:   # noqa: S110
            pass

    def _send(self, sub_id: str, payload: Any) -> bool:
        done = False
        try:
            self.subscribers[sub_id].send(payload)
            done = True
        except (BrokenPipeError, EOFError):
            pass
        return done



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import logging
import argparse
from abc import abstractmethod
from typing import TYPE_CHECKING, Any


try:
    if TYPE_CHECKING:  # pragma: no cover
        from paramiko.channel import Channel

        from ...common.types import HostPort
except ImportError:  # pragma: no cover
    pass

logger = logging.getLogger(__name__)


class BaseSshTunnelHandler:

    def __init__(self, flags: argparse.Namespace) -> None:
        self.flags = flags

    @abstractmethod
    def on_connection(
        self,
        chan: 'Channel',
        origin: 'HostPort',
        server: 'HostPort',
    ) -> None:
        raise NotImplementedError()

    @abstractmethod
    def shutdown(self) -> None:
        raise NotImplementedError()


class BaseSshTunnelListener:

    def __init__(
        self,
        flags: argparse.Namespace,
        handler: BaseSshTunnelHandler,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self.flags = flags
        self.handler = handler

    def __enter__(self) -> 'BaseSshTunnelListener':
        self.setup()
        return self

    def __exit__(self, *args: Any) -> None:
        self.shutdown()

    @abstractmethod
    def is_alive(self) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def is_active(self) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def setup(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def shutdown(self) -> None:
        raise NotImplementedError()



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import struct
from typing import Optional


NULL = b'\x00'


class Socks4Packet:
    """SOCKS4 and SOCKS4a protocol parser.

    FIXME: Currently doesn't buffer during parsing and expects
    packet to arrive within a single socket receive event.
    """

    def __init__(self) -> None:
        # 1 byte, must be equal to 4
        self.vn: Optional[int] = None
        # 1 byte
        self.cd: Optional[int] = None
        # 2 bytes
        self.dstport: Optional[int] = None
        # 4 bytes
        self.dstip: Optional[bytes] = None
        # Variable bytes, NULL terminated
        self.userid: Optional[bytes] = None

    def parse(self, raw: memoryview) -> None:
        cursor = 0
        # Parse vn
        if self.vn is None:
            assert int(raw[cursor]) == 4
            self.vn = 4
        cursor += 1
        # Parse cd
        self.cd = raw[cursor]
        cursor += 1
        # Parse dstport
        self.dstport = struct.unpack('!H', raw[cursor:cursor+2])[0]
        cursor += 2
        # Parse dstip
        self.dstip = struct.unpack('!4s', raw[cursor:cursor+4])[0]
        cursor += 4
        # Parse userid
        ulen = len(raw) - cursor - 1
        self.userid = struct.unpack(
            '!%ds' % ulen, raw[cursor:cursor+ulen],
        )[0]
        cursor += ulen
        # Assert null terminated
        assert raw[cursor] == NULL[0]

    def pack(self) -> bytes:
        user_id = self.userid or b''
        return struct.pack(
            '!bbH4s%ds' % len(user_id),
            self.vn, self.cd,
            self.dstport, self.dstip,
            user_id,
        ) + NULL



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       pre
"""
from .pool import ThreadlessPool
from .work import Work
from .local import BaseLocalExecutor
from .remote import BaseRemoteExecutor
from .delegate import delegate_work_to_pool
from .threaded import start_threaded_work
from .threadless import Threadless


__all__ = [
    'Work',
    'Threadless',
    'ThreadlessPool',
    'delegate_work_to_pool',
    'start_threaded_work',
    'BaseLocalExecutor',
    'BaseRemoteExecutor',
]



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       http
       reusability
"""
import os
import ssl
import time
import errno
import socket
import logging
import threading
import subprocess
from typing import Any, Dict, List, Union, Optional

from .plugin import HttpProxyBasePlugin
from ..parser import HttpParser, httpParserTypes, httpParserStates
from ..plugin import HttpProtocolHandlerPlugin
from ..headers import httpHeaders
from ..methods import httpMethods
from ..exception import HttpProtocolException, ProxyConnectionFailed
from ..protocols import httpProtocols
from ..responses import PROXY_TUNNEL_ESTABLISHED_RESPONSE_PKT
from ...common.pki import gen_csr, sign_csr, gen_public_key
from ...core.event import eventNames
from ...common.flag import flags
from ...common.types import Readables, Writables, Descriptors
from ...common.utils import text_, cert_der_to_dict
from ...core.connection import (
    TcpServerConnection, TcpConnectionUninitializedException,
)
from ...common.constants import (
    COMMA, DEFAULT_CA_FILE, PLUGIN_PROXY_AUTH, DEFAULT_CA_CERT_DIR,
    DEFAULT_CA_KEY_FILE, DEFAULT_CA_CERT_FILE, DEFAULT_DISABLE_HEADERS,
    PROXY_AGENT_HEADER_VALUE, DEFAULT_DISABLE_HTTP_PROXY,
    DEFAULT_CA_SIGNING_KEY_FILE, DEFAULT_INSECURE_TLS_INTERCEPTION,
    DEFAULT_HTTP_PROXY_ACCESS_LOG_FORMAT,
    DEFAULT_HTTPS_PROXY_ACCESS_LOG_FORMAT,
)


logger = logging.getLogger(__name__)


flags.add_argument(
    '--disable-http-proxy',
    action='store_true',
    default=DEFAULT_DISABLE_HTTP_PROXY,
    help='Default: False.  Whether to disable proxy.HttpProxyPlugin.',
)

flags.add_argument(
    '--disable-headers',
    type=str,
    default=COMMA.join(DEFAULT_DISABLE_HEADERS),
    help='Default: None.  Comma separated list of headers to remove before '
    'dispatching client request to upstream server.',
)

flags.add_argument(
    '--ca-key-file',
    type=str,
    default=DEFAULT_CA_KEY_FILE,
    help='Default: None. CA key to use for signing dynamically generated '
    'HTTPS certificates.  If used, must also pass --ca-cert-file and --ca-signing-key-file',
)

flags.add_argument(
    '--insecure-tls-interception',
    action='store_true',
    default=DEFAULT_INSECURE_TLS_INTERCEPTION,
    help='Default: False. Disables certificate verification',
)

flags.add_argument(
    '--ca-cert-dir',
    type=str,
    default=DEFAULT_CA_CERT_DIR,
    help='Default: ~/.proxy/certificates. Directory to store dynamically generated certificates. '
    'Also see --ca-key-file, --ca-cert-file and --ca-signing-key-file',
)

flags.add_argument(
    '--ca-cert-file',
    type=str,
    default=DEFAULT_CA_CERT_FILE,
    help='Default: None. Signing certificate to use for signing dynamically generated '
    'HTTPS certificates.  If used, must also pass --ca-key-file and --ca-signing-key-file',
)

flags.add_argument(
    '--ca-file',
    type=str,
    default=str(DEFAULT_CA_FILE),
    help='Default: ' + str(DEFAULT_CA_FILE) +
    '. Provide path to custom CA bundle for peer certificate verification',
)

flags.add_argument(
    '--ca-signing-key-file',
    type=str,
    default=DEFAULT_CA_SIGNING_KEY_FILE,
    help='Default: None. CA signing key to use for dynamic generation of '
    'HTTPS certificates.  If used, must also pass --ca-key-file and --ca-cert-file',
)

flags.add_argument(
    '--auth-plugin',
    type=str,
    default=PLUGIN_PROXY_AUTH,
    help='Default: ' + PLUGIN_PROXY_AUTH + '.  ' +
    'Auth plugin to use instead of default basic auth plugin.',
)


class HttpProxyPlugin(HttpProtocolHandlerPlugin):
    """HttpProtocolHandler plugin which implements HttpProxy specifications."""

    # Used to synchronization during certificate generation and
    # connection pool operations.
    lock = threading.Lock()

    def __init__(
            self,
            *args: Any, **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.start_time: float = time.time()
        self.upstream: Optional[TcpServerConnection] = None
        self.response: HttpParser = HttpParser(httpParserTypes.RESPONSE_PARSER)
        self.pipeline_request: Optional[HttpParser] = None
        self.pipeline_response: Optional[HttpParser] = None

        self.plugins: Dict[str, HttpProxyBasePlugin] = {}
        if b'HttpProxyBasePlugin' in self.flags.plugins:
            for klass in self.flags.plugins[b'HttpProxyBasePlugin']:
                instance: HttpProxyBasePlugin = klass(
                    self.uid,
                    self.flags,
                    self.client,
                    self.event_queue,
                    self.upstream_conn_pool,
                )
                self.plugins[instance.name()] = instance

    @staticmethod
    def protocols() -> List[int]:
        return [httpProtocols.HTTP_PROXY]

    async def get_descriptors(self) -> Descriptors:
        r: List[int] = []
        w: List[int] = []
        if (
            self.upstream and
            not self.upstream.closed and
            self.upstream.connection
        ):
            r.append(self.upstream.connection.fileno())
        if (
            self.upstream and
            not self.upstream.closed and
            self.upstream.has_buffer() and
            self.upstream.connection
        ):
            w.append(self.upstream.connection.fileno())
        # TODO(abhinavsingh): We need to keep a mapping of plugin and
        # descriptors registered by them, so that within write/read blocks
        # we can invoke the right plugin callbacks.
        for plugin in self.plugins.values():
            plugin_read_desc, plugin_write_desc = await plugin.get_descriptors()
            r.extend(plugin_read_desc)
            w.extend(plugin_write_desc)
        return r, w

    async def write_to_descriptors(self, w: Writables) -> bool:
        if (
            self.upstream
            and not self.upstream.closed
            and self.upstream.connection.fileno() not in w
        ) or not self.upstream:
            # Currently, we just call write/read block of each plugins.  It is
            # plugins responsibility to ignore this callback, if passed descriptors
            # doesn't contain the descriptor they registered.
            for plugin in self.plugins.values():
                teardown = await plugin.write_to_descriptors(w)
                if teardown:
                    return True
        elif self.upstream and not self.upstream.closed and \
                self.upstream.has_buffer() and \
                self.upstream.connection.fileno() in w:
            logger.debug('Server is write ready, flushing...')
            try:
                self.upstream.flush(self.flags.max_sendbuf_size)
            except ssl.SSLWantWriteError:
                logger.warning(
                    'SSLWantWriteError while trying to flush to server, will retry',
                )
                return False
            except BrokenPipeError:
                logger.warning(
                    'BrokenPipeError when flushing buffer for server',
                )
                return self._close_and_release()
            except OSError as e:
                logger.exception(
                    'OSError when flushing buffer to server', exc_info=e,
                )
                return self._close_and_release()
        return False

    async def read_from_descriptors(self, r: Readables) -> bool:
        if (
            self.upstream
            and not self.upstream.closed
            and self.upstream.connection.fileno() not in r
        ) or not self.upstream:
            # Currently, we just call write/read block of each plugins.  It is
            # plugins responsibility to ignore this callback, if passed descriptors
            # doesn't contain the descriptor they registered for.
            for plugin in self.plugins.values():
                teardown = await plugin.read_from_descriptors(r)
                if teardown:
                    return True
        elif self.upstream \
                and not self.upstream.closed \
                and self.upstream.connection.fileno() in r:
            logger.debug('Server is read ready, receiving...')
            try:
                raw = self.upstream.recv(self.flags.server_recvbuf_size)
            except TimeoutError as e:
                self._close_and_release()
                if e.errno == errno.ETIMEDOUT:
                    logger.warning(
                        '%s:%d timed out on recv' %
                        self.upstream.addr,
                    )
                    return True
                raise e
            except ssl.SSLWantReadError:    # Try again later
                # logger.warning('SSLWantReadError encountered while reading from server, will retry ...')
                return False
            except OSError as e:
                if e.errno == errno.EHOSTUNREACH:
                    logger.warning(
                        '%s:%d unreachable on recv' %
                        self.upstream.addr,
                    )
                if e.errno == errno.ECONNRESET:
                    logger.warning(
                        'Connection reset by upstream: {0}:{1}'.format(
                            *self.upstream.addr,
                        ),
                    )
                else:
                    logger.warning(
                        'Exception while receiving from %s connection#%d with reason %r' %
                        (self.upstream.tag, self.upstream.connection.fileno(), e),
                    )
                return self._close_and_release()

            if raw is None:
                logger.debug('Server closed connection, tearing down...')
                return self._close_and_release()

            for plugin in self.plugins.values():
                raw = plugin.handle_upstream_chunk(raw)
                if raw is None:
                    break

            # parse incoming response packet
            # only for non-https requests and when
            # tls interception is enabled
            if raw is not None:
                if not self.request.is_https_tunnel or self._tls_intercept_enabled:
                    if self.response.is_complete:
                        self.handle_pipeline_response(raw)
                    else:
                        self.response.parse(raw)
                        self.emit_response_events(len(raw))
                else:
                    self.response.total_size += len(raw)
                # queue raw data for client
                self.client.queue(raw)
        return False

    def on_client_connection_close(self) -> None:
        context = {
            'client_ip': None if not self.client.addr else self.client.addr[0],
            'client_port': None if not self.client.addr else self.client.addr[1],
            'server_host': text_(self.upstream.addr[0] if self.upstream else None),
            'server_port': text_(self.upstream.addr[1] if self.upstream else None),
            'connection_time_ms': '%.2f' % ((time.time() - self.start_time) * 1000),
            # Request
            'request_method': text_(self.request.method),
            'request_path': text_(self.request.path),
            'request_bytes': text_(self.request.total_size),
            'request_ua': text_(self.request.header(b'user-agent'))
            if self.request.has_header(b'user-agent')
            else None,
            'request_version': text_(self.request.version),
            # Response
            'response_bytes': self.response.total_size,
            'response_code': text_(self.response.code),
            'response_reason': text_(self.response.reason),
        }
        if self.flags.enable_proxy_protocol:
            assert self.request.protocol and self.request.protocol.family
            context.update({
                'protocol': {
                    'family': text_(self.request.protocol.family),
                },
            })
            if self.request.protocol.source:
                context.update({
                    'protocol': {
                        'source_ip': text_(self.request.protocol.source[0]),
                        'source_port': self.request.protocol.source[1],
                    },
                })
            if self.request.protocol.destination:
                context.update({
                    'protocol': {
                        'destination_ip': text_(self.request.protocol.destination[0]),
                        'destination_port': self.request.protocol.destination[1],
                    },
                })

        log_handled = False
        for plugin in self.plugins.values():
            ctx = plugin.on_access_log(context)
            if ctx is None:
                log_handled = True
                break
            context = ctx
        if not log_handled:
            self.access_log(context)

        # Note that, server instance was initialized
        # but not necessarily the connection object exists.
        #
        # Unfortunately this is still being called when an upstream
        # server connection was never established.  This is done currently
        # to assist proxy pool plugin to close its upstream proxy connections.
        #
        # In short, treat on_upstream_connection_close as on_client_connection_close
        # equivalent within proxy plugins.
        #
        # Invoke plugin.on_upstream_connection_close
        for plugin in self.plugins.values():
            plugin.on_upstream_connection_close()

        # If server was never initialized or was _close_and_release
        if self.upstream is None:
            return

        if self.flags.enable_conn_pool:
            assert self.upstream_conn_pool
            # Release the connection for reusability
            with self.lock:
                self.upstream_conn_pool.release(self.upstream)
            return

        try:
            try:
                self.upstream.connection.shutdown(socket.SHUT_WR)
            except OSError:
                pass
            finally:
                # TODO: Unwrap if wrapped before close?
                self.upstream.close()
        except TcpConnectionUninitializedException:
            pass
        finally:
            logger.debug(
                'Closed server connection, has buffer %s' %
                self.upstream.has_buffer(),
            )

    def access_log(self, log_attrs: Dict[str, Any]) -> None:
        access_log_format = DEFAULT_HTTPS_PROXY_ACCESS_LOG_FORMAT
        if not self.request.is_https_tunnel:
            access_log_format = DEFAULT_HTTP_PROXY_ACCESS_LOG_FORMAT
        logger.info(access_log_format.format_map(log_attrs))

    def on_response_chunk(self, chunk: List[memoryview]) -> List[memoryview]:
        # TODO: Allow to output multiple access_log lines
        # for each request over a pipelined HTTP connection (not for HTTPS).
        # However, this must also be accompanied by resetting both request
        # and response objects.
        #
        # if not self.request.is_https_tunnel and \
        #         self.response.is_complete:
        #     self.access_log()
        return chunk

    # Can return None to tear down connection
    def on_client_data(self, raw: memoryview) -> None:
        # For scenarios when an upstream connection was never established,
        # let plugin do whatever they wish to.  These are special scenarios
        # where plugins are trying to do something magical.  Within the core
        # we don't know the context.  In fact, we are not even sure if data
        # exchanged is http spec compliant.
        #
        # Hence, here we pass raw data to HTTP proxy plugins as is.
        #
        # We only call handle_client_data once original request has been
        # completely received
        if not self.upstream:
            for plugin in self.plugins.values():
                o = plugin.handle_client_data(raw)
                if o is None:
                    return
                raw = o
        elif self.upstream and not self.upstream.closed:
            # For http proxy requests, handle pipeline case.
            # We also handle pipeline scenario for https proxy
            # requests is TLS interception is enabled.
            if self.request.is_complete and (
                not self.request.is_https_tunnel or self._tls_intercept_enabled
            ):
                if self.pipeline_request is not None and \
                        self.pipeline_request.is_connection_upgrade:
                    # Previous pipelined request was a WebSocket
                    # upgrade request. Incoming client data now
                    # must be treated as WebSocket protocol packets.
                    self.upstream.queue(raw)
                    return
                if self.pipeline_request is None:
                    # For pipeline requests, we never
                    # want to use --enable-proxy-protocol flag
                    # as proxy protocol header will not be present
                    #
                    # TODO: HTTP parser must be smart about detecting
                    # HA proxy protocol or we must always explicitly pass
                    # the flag when we are expecting HA proxy protocol
                    # request line before HTTP request lines.
                    self.pipeline_request = HttpParser(
                        httpParserTypes.REQUEST_PARSER,
                    )
                self.pipeline_request.parse(raw)
                if self.pipeline_request.is_complete:
                    for plugin in self.plugins.values():
                        assert self.pipeline_request is not None
                        r = plugin.handle_client_request(self.pipeline_request)
                        if r is None:
                            return
                        self.pipeline_request = r
                    assert self.pipeline_request is not None
                    # TODO(abhinavsingh): Remove memoryview wrapping here after
                    # parser is fully memoryview compliant
                    self.upstream.queue(
                        memoryview(
                            self.pipeline_request.build(),
                        ),
                    )
                    if not self.pipeline_request.is_connection_upgrade:
                        self.pipeline_request = None
            # For scenarios where we cannot peek into the data,
            # simply queue for upstream server.
            else:
                self.upstream.queue(raw)

    @property
    def _tls_intercept_enabled(self) -> bool:
        do_intercept = self.tls_interception_enabled
        if not do_intercept:
            return do_intercept
        # If enabled by flags, check if a plugin wants us to bypass
        # interception for this particular request
        for plugin in self.plugins.values():
            do_intercept = plugin.do_intercept(self.request)
            # A plugin requested to not intercept the request
            if do_intercept is False:
                break
        return do_intercept

    def on_request_complete(self) -> Union[socket.socket, bool]:
        self.emit_request_complete()

        # Invoke plugin.before_upstream_connection
        #
        # before_upstream_connection can:
        # 1) Raise HttpRequestRejected exception to reject the connection
        # 2) return None to continue without establishing an upstream server connection
        #    e.g. for scenarios when plugins want to return response from cache, or,
        #    via out-of-band over the network request.
        do_connect = True
        for plugin in self.plugins.values():
            r = plugin.before_upstream_connection(self.request)
            if r is None:
                do_connect = False
                break
            self.request = r

        # Connect to upstream
        if do_connect:
            self.connect_upstream()

        # Invoke plugin.handle_client_request
        for plugin in self.plugins.values():
            assert self.request is not None
            r = plugin.handle_client_request(self.request)
            if r is not None:
                self.request = r
            else:
                return False

        # For https requests, respond back with tunnel established response.
        # Optionally, setup interceptor if TLS interception is enabled.
        if self.upstream:
            if self.request.is_https_tunnel:
                self.client.queue(PROXY_TUNNEL_ESTABLISHED_RESPONSE_PKT)
                if self._tls_intercept_enabled:
                    return self.intercept()
            # If an upstream server connection was established for http request,
            # queue the request for upstream server.
            else:
                # - proxy-connection header is a mistake, it doesn't seem to be
                #   officially documented in any specification, drop it.
                # - proxy-authorization is of no use for upstream, remove it.
                self.request.del_headers(
                    [
                        httpHeaders.PROXY_AUTHORIZATION,
                        httpHeaders.PROXY_CONNECTION,
                    ],
                )
                # - For HTTP/1.0, connection header defaults to close
                # - For HTTP/1.1, connection header defaults to keep-alive
                # Respect headers sent by client instead of manipulating
                # Connection or Keep-Alive header.  However, note that per
                # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Connection
                # connection headers are meant for communication between client and
                # first intercepting proxy.
                self.request.add_headers(
                    [(b'Via', b'1.1 %s' % PROXY_AGENT_HEADER_VALUE)],
                )
                # Disable args.disable_headers before dispatching to upstream
                self.upstream.queue(
                    memoryview(
                        self.request.build(
                            disable_headers=self.flags.disable_headers,
                        ),
                    ),
                )
        return False

    def handle_pipeline_response(self, raw: memoryview) -> None:
        if self.pipeline_response is None:
            self.pipeline_response = HttpParser(
                httpParserTypes.RESPONSE_PARSER,
            )
        self.pipeline_response.parse(raw)
        if self.pipeline_response.is_complete:
            self.pipeline_response = None

    def connect_upstream(self) -> None:
        host, port = self.request.host, self.request.port
        if host and port:
            try:
                # Invoke plugin.resolve_dns
                upstream_ip, source_addr = None, None
                for plugin in self.plugins.values():
                    upstream_ip, source_addr = plugin.resolve_dns(
                        text_(host), port,
                    )
                    if upstream_ip or source_addr:
                        break
                logger.debug(
                    'Connecting to upstream %s:%d' %
                    (text_(host), port),
                )
                if self.flags.enable_conn_pool:
                    assert self.upstream_conn_pool
                    with self.lock:
                        created, self.upstream = self.upstream_conn_pool.acquire(
                            (text_(host), port),
                        )
                else:
                    created, self.upstream = True, TcpServerConnection(
                        text_(host), port,
                    )
                    # Connect with overridden upstream IP and source address
                    # if any of the plugin returned a non-null value.
                    self.upstream.connect(
                        addr=None if not upstream_ip else (
                            upstream_ip, port,
                        ), source_address=source_addr,
                    )
                    self.upstream.connection.setblocking(False)
                if not created:
                    # NOTE: Acquired connection might be in an unusable state.
                    #
                    # This can only be confirmed by reading from connection.
                    # For stale connections, we will receive None, indicating
                    # to drop the connection.
                    #
                    # If that happen, we must acquire a fresh connection.
                    logger.info(
                        'Reusing connection to upstream %s:%d' %
                        (text_(host), port),
                    )
                    return
                logger.debug(
                    'Connected to upstream %s:%s' %
                    (text_(host), port),
                )
            except Exception as e:  # TimeoutError, socket.gaierror
                logger.warning(
                    'Unable to connect with upstream %s:%d due to %s' % (
                        text_(host), port, str(e),
                    ),
                )
                if self.flags.enable_conn_pool and self.upstream:
                    assert self.upstream_conn_pool
                    with self.lock:
                        self.upstream_conn_pool.release(self.upstream)
                raise ProxyConnectionFailed(
                    text_(host), port, repr(e),
                ) from e
        else:
            raise HttpProtocolException('Both host and port must exist')

    #
    # Interceptor related methods
    #

    def gen_ca_signed_certificate(
            self, cert_file_path: str, certificate: Dict[str, Any],
    ) -> None:
        '''CA signing key (default) is used for generating a public key
        for common_name, if one already doesn't exist.  Using generated
        public key a CSR request is generated, which is then signed by
        CA key and secret.  Again this process only happen if signed
        certificate doesn't already exist.

        returns signed certificate path.'''
        assert(
            self.request.host and self.flags.ca_cert_dir and self.flags.ca_signing_key_file and
            self.flags.ca_key_file and self.flags.ca_cert_file
        )

        upstream_subject = {s[0][0]: s[0][1] for s in certificate['subject']}
        public_key_path = os.path.join(
            self.flags.ca_cert_dir,
            '{0}.{1}'.format(text_(self.request.host), 'pub'),
        )
        private_key_path = self.flags.ca_signing_key_file
        private_key_password = ''

        # Build certificate subject
        keys = {
            'CN': 'commonName',
            'C': 'countryName',
            'ST': 'stateOrProvinceName',
            'L': 'localityName',
            'O': 'organizationName',
            'OU': 'organizationalUnitName',
        }
        subject = ''
        for key in keys:
            if upstream_subject.get(keys[key], None):
                subject += '/{0}={1}'.format(
                    key,
                    upstream_subject.get(keys[key]),
                )
        alt_subj_names = [text_(self.request.host)]
        validity_in_days = 365 * 2
        timeout = 10

        # Generate a public key for the common name
        if not os.path.isfile(public_key_path):
            logger.debug('Generating public key %s', public_key_path)
            resp = gen_public_key(
                public_key_path=public_key_path, private_key_path=private_key_path,
                private_key_password=private_key_password, subject=subject, alt_subj_names=alt_subj_names,
                validity_in_days=validity_in_days, timeout=timeout,
                openssl=self.flags.openssl,
            )
            assert(resp is True)

        csr_path = os.path.join(
            self.flags.ca_cert_dir,
            '{0}.{1}'.format(text_(self.request.host), 'csr'),
        )

        # Generate a CSR request for this common name
        if not os.path.isfile(csr_path):
            logger.debug('Generating CSR %s', csr_path)
            resp = gen_csr(
                csr_path=csr_path, key_path=private_key_path, password=private_key_password,
                crt_path=public_key_path, timeout=timeout,
                openssl=self.flags.openssl,
            )
            assert(resp is True)

        ca_key_path = self.flags.ca_key_file
        ca_key_password = ''
        ca_crt_path = self.flags.ca_cert_file
        serial = '%d%d' % (time.time(), os.getpid())

        # Sign generated CSR
        if not os.path.isfile(cert_file_path):
            logger.debug('Signing CSR %s', cert_file_path)
            resp = sign_csr(
                csr_path=csr_path, crt_path=cert_file_path, ca_key_path=ca_key_path,
                ca_key_password=ca_key_password, ca_crt_path=ca_crt_path,
                serial=str(serial), alt_subj_names=alt_subj_names,
                validity_in_days=validity_in_days, timeout=timeout,
                openssl=self.flags.openssl,
            )
            assert(resp is True)

    @staticmethod
    def generated_cert_file_path(ca_cert_dir: str, host: str) -> str:
        return os.path.join(ca_cert_dir, '%s.pem' % host)

    def generate_upstream_certificate(
            self, certificate: Dict[str, Any],
    ) -> str:
        if not (
            self.flags.ca_cert_dir and self.flags.ca_signing_key_file and
            self.flags.ca_cert_file and self.flags.ca_key_file
        ):
            raise HttpProtocolException(
                f'For certificate generation all the following flags are mandatory: '
                f'--ca-cert-file:{ self.flags.ca_cert_file}, '
                f'--ca-key-file:{ self.flags.ca_key_file}, '
                f'--ca-signing-key-file:{ self.flags.ca_signing_key_file}',
            )
        cert_file_path = HttpProxyPlugin.generated_cert_file_path(
            self.flags.ca_cert_dir, text_(self.request.host),
        )
        with self.lock:
            if not os.path.isfile(cert_file_path):
                self.gen_ca_signed_certificate(cert_file_path, certificate)
        return cert_file_path

    def intercept(self) -> Union[socket.socket, bool]:
        # Perform SSL/TLS handshake with upstream
        teardown = self.wrap_server()
        if teardown:
            return teardown
        # Generate certificate and perform handshake with client
        # wrap_client also flushes client data before wrapping
        # sending to client can raise, handle expected exceptions
        teardown = self.wrap_client()
        if teardown:
            return teardown
        # Update all plugin connection reference
        # TODO(abhinavsingh): Is this required?
        for plugin in self.plugins.values():
            plugin.client._conn = self.client.connection
        return self.client.connection

    def wrap_server(self) -> bool:
        assert self.upstream is not None
        assert isinstance(self.upstream.connection, socket.socket)
        do_close = False
        try:
            # pylint: disable=E1101
            verify_mode = (
                ssl.VerifyMode.CERT_NONE
                if self.flags.insecure_tls_interception
                else ssl.VerifyMode.CERT_REQUIRED
            )
            self.upstream.wrap(
                text_(self.request.host),
                self.flags.ca_file,
                as_non_blocking=True,
                verify_mode=verify_mode,
            )
        except ssl.SSLCertVerificationError:    # Server raised certificate verification error
            # When --disable-interception-on-ssl-cert-verification-error flag is on,
            # we will cache such upstream hosts and avoid intercepting them for future
            # requests.
            logger.warning(
                'ssl.SSLCertVerificationError: ' +
                'Server raised cert verification error for upstream: {0}'.format(
                    self.upstream.addr[0],
                ),
            )
            do_close = True
        except ssl.SSLError as e:
            if e.reason == 'SSLV3_ALERT_HANDSHAKE_FAILURE':
                logger.warning(
                    '{0}: '.format(e.reason) +
                    'Server raised handshake alert failure for upstream: {0}'.format(
                        self.upstream.addr[0],
                    ),
                )
            else:
                logger.exception(
                    'SSLError when wrapping client for upstream: {0}'.format(
                        self.upstream.addr[0],
                    ), exc_info=e,
                )
            do_close = True
        if not do_close:
            assert isinstance(self.upstream.connection, ssl.SSLSocket)
        return do_close

    def wrap_client(self) -> bool:
        assert self.upstream is not None and self.flags.ca_signing_key_file is not None
        assert isinstance(self.upstream.connection, ssl.SSLSocket)
        do_close = False
        try:
            # TODO: Perform async certificate generation
            generated_cert = self.generate_upstream_certificate(
                cert_der_to_dict(self.upstream.connection.getpeercert(True)),
            )
            self.client.wrap(self.flags.ca_signing_key_file, generated_cert)
        except subprocess.TimeoutExpired as e:  # Popen communicate timeout
            logger.exception(
                'TimeoutExpired during certificate generation', exc_info=e,
            )
            do_close = True
        except ssl.SSLCertVerificationError:    # Client raised certificate verification error
            # When --disable-interception-on-ssl-cert-verification-error flag is on,
            # we will cache such upstream hosts and avoid intercepting them for future
            # requests.
            logger.warning(
                'ssl.SSLCertVerificationError: ' +
                'Client raised cert verification error for upstream: {0}'.format(
                    self.upstream.addr[0],
                ),
            )
            do_close = True
        except ssl.SSLEOFError as e:
            logger.warning(
                'ssl.SSLEOFError {0} when wrapping client for upstream: {1}'.format(
                    str(e), self.upstream.addr[0],
                ),
            )
            do_close = True
        except ssl.SSLError as e:
            if e.reason in ('TLSV1_ALERT_UNKNOWN_CA', 'UNSUPPORTED_PROTOCOL'):
                logger.warning(
                    '{0}: '.format(e.reason) +
                    'Client raised cert verification error for upstream: {0}'.format(
                        self.upstream.addr[0],
                    ),
                )
            else:
                logger.exception(
                    'OSError when wrapping client for upstream: {0}'.format(
                        self.upstream.addr[0],
                    ), exc_info=e,
                )
            do_close = True
        except BrokenPipeError:
            logger.warning(
                'BrokenPipeError when wrapping client for upstream: {0}'.format(
                    self.upstream.addr[0],
                ),
            )
            do_close = True
        except OSError as e:
            logger.exception(
                'OSError when wrapping client for upstream: {0}'.format(
                    self.upstream.addr[0],
                ), exc_info=e,
            )
            do_close = True
        if not do_close:
            logger.debug('TLS intercepting using %s', generated_cert)
        return do_close

    #
    # Event emitter callbacks
    #

    def emit_request_complete(self) -> None:
        if not self.flags.enable_events:
            return
        assert self.request.port and self.event_queue
        self.event_queue.publish(
            request_id=self.uid,
            event_name=eventNames.REQUEST_COMPLETE,
            event_payload={
                'url': text_(self.request.path)
                if self.request.is_https_tunnel
                else 'http://%s:%d%s' % (text_(self.request.host), self.request.port, text_(self.request.path)),
                'method': text_(self.request.method),
                'headers': {}
                if not self.request.headers else
                {
                    text_(k): text_(v[1])
                    for k, v in self.request.headers.items()
                },
                'body': text_(self.request.body, errors='ignore')
                if self.request.method == httpMethods.POST
                else None,
            },
            publisher_id=self.__class__.__qualname__,
        )

    def emit_response_events(self, chunk_size: int) -> None:
        if not self.flags.enable_events:
            return
        if self.response.is_complete:
            self.emit_response_complete()
        elif self.response.state == httpParserStates.RCVING_BODY:
            self.emit_response_chunk_received(chunk_size)
        elif self.response.state == httpParserStates.HEADERS_COMPLETE:
            self.emit_response_headers_complete()

    def emit_response_headers_complete(self) -> None:
        if not self.flags.enable_events:
            return
        assert self.event_queue
        self.event_queue.publish(
            request_id=self.uid,
            event_name=eventNames.RESPONSE_HEADERS_COMPLETE,
            event_payload={
                'headers': {}
                if not self.response.headers else
                {
                    text_(k): text_(v[1])
                    for k, v in self.response.headers.items()
                },
            },
            publisher_id=self.__class__.__qualname__,
        )

    def emit_response_chunk_received(self, chunk_size: int) -> None:
        if not self.flags.enable_events:
            return
        assert self.event_queue
        self.event_queue.publish(
            request_id=self.uid,
            event_name=eventNames.RESPONSE_CHUNK_RECEIVED,
            event_payload={
                'chunk_size': chunk_size,
                'encoded_chunk_size': chunk_size,
            },
            publisher_id=self.__class__.__qualname__,
        )

    def emit_response_complete(self) -> None:
        if not self.flags.enable_events:
            return
        assert self.event_queue
        self.event_queue.publish(
            request_id=self.uid,
            event_name=eventNames.RESPONSE_COMPLETE,
            event_payload={
                'encoded_response_size': self.response.total_size,
            },
            publisher_id=self.__class__.__qualname__,
        )

    #
    # Internal methods
    #

    def _close_and_release(self) -> bool:
        if self.flags.enable_conn_pool:
            assert self.upstream and not self.upstream.closed and self.upstream_conn_pool
            self.upstream.closed = True
            with self.lock:
                self.upstream_conn_pool.release(self.upstream)
            self.upstream = None
        return True



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       http
       iterable
"""
from typing import NamedTuple


# Ref: https://www.iana.org/assignments/http-methods/http-methods.xhtml
HttpMethods = NamedTuple(
    'HttpMethods', [
        ('ACL', bytes),
        ('BASELINE_CONTROL', bytes),
        ('BIND', bytes),
        ('CHECKIN', bytes),
        ('CHECKOUT', bytes),
        ('CONNECT', bytes),
        ('COPY', bytes),
        ('DELETE', bytes),
        ('GET', bytes),
        ('HEAD', bytes),
        ('LABEL', bytes),
        ('LINK', bytes),
        ('LOCK', bytes),
        ('MERGE', bytes),
        ('MKACTIVITY', bytes),
        ('MKCALENDAR', bytes),
        ('MKCOL', bytes),
        ('MKREDIRECTREF', bytes),
        ('MKWORKSPACE', bytes),
        ('MOVE', bytes),
        ('OPTIONS', bytes),
        ('ORDERPATCH', bytes),
        ('PATCH', bytes),
        ('POST', bytes),
        ('PRI', bytes),
        ('PROPFIND', bytes),
        ('PROPPATCH', bytes),
        ('PUT', bytes),
        ('REBIND', bytes),
        ('REPORT', bytes),
        ('SEARCH', bytes),
        ('TRACE', bytes),
        ('UNBIND', bytes),
        ('UNCHECKOUT', bytes),
        ('UNLINK', bytes),
        ('UNLOCK', bytes),
        ('UPDATE', bytes),
        ('UPDATEREDIRECTREF', bytes),
        ('VERSION_CONTROL', bytes),
        ('STAR', bytes),
    ],
)

httpMethods = HttpMethods(
    b'ACL',
    b'BASELINE-CONTROL',
    b'BIND',
    b'CHECKIN',
    b'CHECKOUT',
    b'CONNECT',
    b'COPY',
    b'DELETE',
    b'GET',
    b'HEAD',
    b'LABEL',
    b'LINK',
    b'LOCK',
    b'MERGE',
    b'MKACTIVITY',
    b'MKCALENDAR',
    b'MKCOL',
    b'MKREDIRECTREF',
    b'MKWORKSPACE',
    b'MOVE',
    b'OPTIONS',
    b'ORDERPATCH',
    b'PATCH',
    b'POST',
    b'PRI',
    b'PROPFIND',
    b'PROPPATCH',
    b'PUT',
    b'REBIND',
    b'REPORT',
    b'SEARCH',
    b'TRACE',
    b'UNBIND',
    b'UNCHECKOUT',
    b'UNLINK',
    b'UNLOCK',
    b'UPDATE',
    b'UPDATEREDIRECTREF',
    b'VERSION-CONTROL',
    b'*',
)



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       reusability
       Submodules
"""
from .pool import UpstreamConnectionPool
from .types import tcpConnectionTypes
from .client import TcpClientConnection
from .server import TcpServerConnection
from .connection import TcpConnection, TcpConnectionUninitializedException


__all__ = [
    'TcpConnection',
    'TcpConnectionUninitializedException',
    'TcpServerConnection',
    'TcpClientConnection',
    'tcpConnectionTypes',
    'UpstreamConnectionPool',
]



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import ssl
import logging
from abc import ABC, abstractmethod
from typing import Any, Optional

from ...common.types import Readables, Writables, Descriptors
from ...core.connection import TcpServerConnection


logger = logging.getLogger(__name__)


class TcpUpstreamConnectionHandler(ABC):
    """:class:`~proxy.core.base.TcpUpstreamConnectionHandler` can
    be used to insert an upstream server connection lifecycle.

    Call `initialize_upstream` to initialize the upstream connection object.
    Then, directly use ``self.upstream`` object within your class.

    See :class:`~proxy.plugin.proxy_pool.ProxyPoolPlugin` for example usage.
    """

    def __init__(self, *args: Any,  **kwargs: Any) -> None:
        # This is currently a hack, see comments below for rationale,
        # will be fixed later.
        super().__init__(*args, **kwargs)   # type: ignore
        self.upstream: Optional[TcpServerConnection] = None
        # TODO: Currently, :class:`~proxy.core.base.TcpUpstreamConnectionHandler`
        # is used within :class:`~proxy.http.server.ReverseProxy` and
        # :class:`~proxy.plugin.ProxyPoolPlugin`.
        #
        # For both of which we expect a 4-tuple as arguments
        # containing (uuid, flags, client, event_queue).
        # We really don't need the rest of the args here.
        # May be uuid?  May be event_queue in the future.
        # But certainly we don't not client here.
        # A separate tunnel class must be created which handles
        # client connection too.
        #
        # Both :class:`~proxy.http.server.ReverseProxy` and
        # :class:`~proxy.plugin.ProxyPoolPlugin` are currently
        # calling client queue within `handle_upstream_data` callback.
        #
        # This can be abstracted out too.
        self.server_recvbuf_size = args[1].server_recvbuf_size
        self.total_size = 0

    @abstractmethod
    def handle_upstream_data(self, raw: memoryview) -> None:
        raise NotImplementedError()     # pragma: no cover

    def initialize_upstream(self, addr: str, port: int) -> None:
        self.upstream = TcpServerConnection(addr, port)

    async def get_descriptors(self) -> Descriptors:
        if not self.upstream:
            return [], []
        return [self.upstream.connection.fileno()], \
            [self.upstream.connection.fileno()] \
            if self.upstream.has_buffer() \
            else []

    async def read_from_descriptors(self, r: Readables) -> bool:
        if self.upstream and \
                self.upstream.connection.fileno() in r:
            try:
                raw = self.upstream.recv(self.server_recvbuf_size)
                if raw is None:     # pragma: no cover
                    # Tear down because upstream proxy closed the connection
                    return True
                self.total_size += len(raw)
                self.handle_upstream_data(raw)
            except TimeoutError:    # pragma: no cover
                logger.info('Upstream recv timeout error')
                return True
            except ssl.SSLWantReadError:    # pragma: no cover
                logger.info('Upstream SSLWantReadError, will retry')
                return False
            except ConnectionResetError:    # pragma: no cover
                logger.debug('Connection reset by upstream')
                return True
        return False

    async def write_to_descriptors(self, w: Writables) -> bool:
        if self.upstream and \
                self.upstream.connection.fileno() in w and \
                self.upstream.has_buffer():
            try:
                # TODO: max sendbuf size flag currently not used here
                self.upstream.flush()
            except ssl.SSLWantWriteError:   # pragma: no cover
                logger.info('Upstream SSLWantWriteError, will retry')
                return False
            except BrokenPipeError:     # pragma: no cover
                logger.debug('BrokenPipeError when flushing to upstream')
                return True
        return False



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       auth
       http
"""
from typing import Optional

from ...http import httpHeaders
from ..exception import ProxyAuthenticationFailed
from ...http.proxy import HttpProxyBasePlugin
from ...http.parser import HttpParser


class AuthPlugin(HttpProxyBasePlugin):
    """Performs proxy authentication."""

    def before_upstream_connection(
            self, request: HttpParser,
    ) -> Optional[HttpParser]:
        if self.flags.auth_code:
            request.headers = request.headers or {}
            if httpHeaders.PROXY_AUTHORIZATION not in request.headers:
                raise ProxyAuthenticationFailed()
            parts = request.headers[httpHeaders.PROXY_AUTHORIZATION][1].split()
            if len(parts) != 2 \
                    or parts[0].lower() != b'basic' \
                    or parts[1] != self.flags.auth_code:
                raise ProxyAuthenticationFailed()
        return request



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       Lua
"""
import re
from typing import TYPE_CHECKING, List, Tuple, Union

from ..http import Url
from ..http.parser import HttpParser
from ..http.server import ReverseProxyBasePlugin
from ..common.types import RePattern
from ..http.exception.base import HttpProtocolException


if TYPE_CHECKING:
    from ..core.connection import TcpServerConnection


class ReverseProxyPlugin(ReverseProxyBasePlugin):
    """This example plugin is equivalent to following Nginx configuration::

        ```text
        location /get {
            proxy_pass http://httpbin.org/get
        }
        ```

    Plugin also demonstrates how to write "Python" equivalent for any
    "Nginx Lua" based configuration i.e. your plugin code will have
    full control over what do after one of your route has matched.
    """

    def routes(self) -> List[Union[str, Tuple[str, List[bytes]]]]:
        return [
            # A static route
            (
                r'/get$',
                [b'http://httpbingo.org/get', b'https://httpbingo.org/get'],
            ),
            # A dynamic route to catch requests on "/get/<int>""
            # See "handle_route" method below for what we do when
            # this pattern matches.
            r'/get/(\d+)$',
        ]

    def handle_route(
        self,
        request: HttpParser,
        pattern: RePattern,
    ) -> Union[memoryview, Url, 'TcpServerConnection']:
        """For our example dynamic route, we want to simply convert
        any incoming request to "/get/1" into "/get?id=1" when serving from upstream.
        """
        choice: Url = Url.from_bytes(b'http://httpbingo.org/get')
        assert request.path
        result = re.search(pattern, request.path.decode())
        if not result or len(result.groups()) != 1:
            raise HttpProtocolException('Invalid request')
        assert choice.remainder == b'/get'
        # NOTE: Internally, reverse proxy core replaces
        # original request.path with the choice.remainder value.
        # e.g. for this example, request.path will be "/get/1".
        # Core will automatically replace that with "/get?id=1"
        # before dispatching request to choice of upstream server.
        choice.remainder += f'?id={result.groups()[0]}'.encode()
        return choice



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""


from typing import Tuple, Optional

from proxy.proxy import GroutClientBasePlugin
from proxy.common.types import HostPort
from proxy.http.parser.parser import HttpParser


class GroutClientPlugin(GroutClientBasePlugin):

    def resolve_route(
        self,
        route: str,
        request: HttpParser,
        origin: HostPort,
        server: HostPort,
    ) -> Tuple[Optional[str], HttpParser]:
        # print(request, origin, server, '->', route)
        # print(request.header(b'host'), request.path)
        #
        # Send to localhost:7001 irrespective of the
        # original "route" value provided to the grout client
        # OR any custom host:upstream mapping provided through the
        # --tunnel-route flags.
        #
        # Optionally, you can also strip path like this:
        # request.path = b"/"
        #
        # Return None for route to drop the request
        # return None, request
        #
        return 'http://localhost:7001', request



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import time


class Leakage:
    """Leaky Bucket algorithm."""

    def __init__(self, rate: int) -> None:
        """Initialize the leaky bucket with a specified leak rate in bytes per second."""
        # Maximum number of tokens the bucket can hold (bytes per second)
        self.rate = rate
        self.tokens = rate
        self.last_check = time.time()

    def _refill(self) -> None:
        """Refill tokens based on the elapsed time since the last check."""
        now = time.time()
        elapsed = now - self.last_check
        # Add tokens proportional to elapsed time, up to the rate
        self.tokens += int(elapsed * self.rate)
        # Cap tokens at the maximum rate to enforce the rate limit
        self.tokens = min(self.tokens, self.rate)
        self.last_check = now

    def release(self, tokens: int) -> None:
        """When you are unable to consume amount units of token, release them into the bucket.

        E.g. say you wanted to read 1024 units, but only 24 units were read, then put
        back unconsumed 1000 tokens back in the bucket."""
        if tokens < 0:
            raise ValueError('Cannot release a negative number of tokens')
        self.tokens += tokens
        self.tokens = min(self.tokens, self.rate)

    def consume(self, amount: int) -> int:
        """Attempt to consume the amount from the bucket.

        Returns the amount allowed to be sent, up to the available tokens (rate).
        """
        self._refill()
        allowed = min(amount, self.tokens)
        self.tokens -= allowed
        return allowed



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import re
import time
import socket
import logging
from typing import Any, Dict, List, Tuple, Union, Pattern, Optional

from .plugin import HttpWebServerBasePlugin
from ..parser import HttpParser, httpParserTypes
from ..plugin import HttpProtocolHandlerPlugin
from ..methods import httpMethods
from .protocols import httpProtocolTypes
from ..exception import HttpProtocolException
from ..protocols import httpProtocols
from ..responses import NOT_FOUND_RESPONSE_PKT
from ..websocket import WebsocketFrame, websocketOpcodes
from ...core.event import eventNames
from ...common.flag import flags
from ...common.types import Readables, Writables, Descriptors
from ...common.utils import text_, build_websocket_handshake_response
from ...common.constants import (
    DEFAULT_ENABLE_WEB_SERVER, DEFAULT_STATIC_SERVER_DIR,
    DEFAULT_ENABLE_REWRITE_HOST, DEFAULT_ENABLE_REVERSE_PROXY,
    DEFAULT_ENABLE_STATIC_SERVER, DEFAULT_WEB_ACCESS_LOG_FORMAT,
    DEFAULT_MIN_COMPRESSION_LENGTH,
)


logger = logging.getLogger(__name__)


flags.add_argument(
    '--enable-web-server',
    action='store_true',
    default=DEFAULT_ENABLE_WEB_SERVER,
    help='Default: False.  Whether to enable proxy.HttpWebServerPlugin.',
)

flags.add_argument(
    '--enable-static-server',
    action='store_true',
    default=DEFAULT_ENABLE_STATIC_SERVER,
    help='Default: False.  Enable inbuilt static file server. '
    'Optionally, also use --static-server-dir to serve static content '
    'from custom directory.  By default, static file server serves '
    'out of installed proxy.py python module folder.',
)

flags.add_argument(
    '--static-server-dir',
    type=str,
    default=DEFAULT_STATIC_SERVER_DIR,
    help='Default: "public" folder in directory where proxy.py is placed. '
    'This option is only applicable when static server is also enabled. '
    'See --enable-static-server.',
)

flags.add_argument(
    '--min-compression-length',
    type=int,
    default=DEFAULT_MIN_COMPRESSION_LENGTH,
    help='Default: ' + str(DEFAULT_MIN_COMPRESSION_LENGTH) + ' bytes.  ' +
    'Sets the minimum length of a response that will be compressed (gzipped).',
)

flags.add_argument(
    '--enable-reverse-proxy',
    action='store_true',
    default=DEFAULT_ENABLE_REVERSE_PROXY,
    help='Default: False.  Whether to enable reverse proxy core.',
)

flags.add_argument(
    '--rewrite-host-header',
    action='store_true',
    default=DEFAULT_ENABLE_REWRITE_HOST,
    help='Default: '
    + str(DEFAULT_ENABLE_REWRITE_HOST)
    + '.  '
    + 'If used, reverse proxy server will rewrite Host header field before sending to upstream.',
)


class HttpWebServerPlugin(HttpProtocolHandlerPlugin):
    """HttpProtocolHandler plugin which handles incoming requests to local web server."""

    def __init__(
            self,
            *args: Any, **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.start_time: float = time.time()
        self.pipeline_request: Optional[HttpParser] = None
        self.switched_protocol: Optional[int] = None
        self.route: Optional[HttpWebServerBasePlugin] = None

        self.plugins: Dict[str, HttpWebServerBasePlugin] = {}
        self.routes: Dict[
            int, Dict[Pattern[str], HttpWebServerBasePlugin],
        ] = {
            httpProtocolTypes.HTTP: {},
            httpProtocolTypes.HTTPS: {},
            httpProtocolTypes.WEBSOCKET: {},
        }
        if b'HttpWebServerBasePlugin' in self.flags.plugins:
            self._initialize_web_plugins()

        self._response_size = 0
        self._post_request_data_size = 0

    @staticmethod
    def protocols() -> List[int]:
        return [httpProtocols.WEB_SERVER]

    def _initialize_web_plugins(self) -> None:
        for klass in self.flags.plugins[b'HttpWebServerBasePlugin']:
            instance: HttpWebServerBasePlugin = klass(
                self.uid,
                self.flags,
                self.client,
                self.event_queue,
                self.upstream_conn_pool,
            )
            self.plugins[instance.name()] = instance
            for (protocol, route) in instance.routes():
                pattern = re.compile(route)
                self.routes[protocol][pattern] = self.plugins[instance.name()]

    def encryption_enabled(self) -> bool:
        return self.flags.keyfile is not None and \
            self.flags.certfile is not None

    def switch_to_websocket(self) -> None:
        self.client.queue(
            memoryview(
                build_websocket_handshake_response(
                    WebsocketFrame.key_to_accept(
                        self.request.header(b'Sec-WebSocket-Key'),
                    ),
                ),
            ),
        )
        self.switched_protocol = httpProtocolTypes.WEBSOCKET

    def on_request_complete(self) -> Union[socket.socket, bool]:
        self.emit_request_complete()
        path = self.request.path or b'/'
        teardown = self._try_route(path)
        if teardown:
            return teardown
        # No-route found, try static serving if enabled
        if self.route is None:
            if self.flags.enable_static_server:
                self._try_static_or_404(path)
                return True
            # Catch all unhandled web server requests, return 404
            self.client.queue(NOT_FOUND_RESPONSE_PKT)
            return True
        return False

    async def get_descriptors(self) -> Descriptors:
        r, w = [], []
        for plugin in self.plugins.values():
            r1, w1 = await plugin.get_descriptors()
            r.extend(r1)
            w.extend(w1)
        return r, w

    async def write_to_descriptors(self, w: Writables) -> bool:
        for plugin in self.plugins.values():
            teardown = await plugin.write_to_descriptors(w)
            if teardown:
                return True
        return False

    async def read_from_descriptors(self, r: Readables) -> bool:
        for plugin in self.plugins.values():
            teardown = await plugin.read_from_descriptors(r)
            if teardown:
                return True
        return False

    def on_client_data(self, raw: memoryview) -> None:
        self._post_request_data_size += len(raw)
        if self.route and self.route.on_client_data(self.request, raw) is None:
            return
        if self.switched_protocol == httpProtocolTypes.WEBSOCKET:
            # TODO(abhinavsingh): Do we really tobytes() here?
            # Websocket parser currently doesn't depend on internal
            # buffers, due to which it can directly parse out of
            # memory views.  But how about large payloads scenarios?
            remaining = raw.tobytes()
            frame = WebsocketFrame()
            while remaining != b'':
                # TODO: Tear down if invalid protocol exception
                remaining = frame.parse(remaining)
                if frame.opcode == websocketOpcodes.CONNECTION_CLOSE:
                    raise HttpProtocolException(
                        'Client sent connection close packet',
                    )
                else:
                    assert self.route
                    self.route.on_websocket_message(frame)
                frame.reset()
            return
        # If 1st valid request was completed and it's a HTTP/1.1 keep-alive
        # And only if we have a route, parse pipeline requests
        if self.request.is_complete and \
                self.request.is_http_1_1_keep_alive and \
                self.route is not None:
            if self.pipeline_request is None:
                self.pipeline_request = HttpParser(
                    httpParserTypes.REQUEST_PARSER,
                )
            self.pipeline_request.parse(raw)
            if self.pipeline_request.is_complete:
                self.route.handle_request(self.pipeline_request)
                if not self.pipeline_request.is_http_1_1_keep_alive:
                    raise HttpProtocolException(
                        'Pipelined request is not keep-alive, will tear down request...',
                    )
                self.pipeline_request = None

    def on_response_chunk(self, chunk: List[memoryview]) -> List[memoryview]:
        self._response_size += sum(len(c) for c in chunk)
        return chunk

    def _context(self) -> Dict[str, Any]:
        return {
            'client_ip': None if not self.client.addr else self.client.addr[0],
            'client_port': None if not self.client.addr else self.client.addr[1],
            'connection_time_ms': '%.2f' % ((time.time() - self.start_time) * 1000),
            # Request
            'request_method': text_(self.request.method),
            'request_path': text_(self.request.path),
            'request_bytes': self.request.total_size + self._post_request_data_size,
            'request_ua': (
                text_(self.request.header(b'user-agent'))
                if self.request.has_header(b'user-agent')
                else None
            ),
            'request_version': (
                None if not self.request.version else text_(self.request.version)
            ),
            # Response
            #
            # TODO: Track and inject web server specific response attributes
            # Currently, plugins are allowed to queue raw bytes, because of
            # which we'll have to reparse the queued packets to deduce
            # several attributes required below.  At least for code and
            # reason attributes.
            #
            'response_bytes': self._response_size,
            # 'response_code': text_(self.response.code),
            # 'response_reason': text_(self.response.reason),
        }

    def on_client_connection_close(self) -> None:
        context = self._context()
        log_handled = False
        if self.route:
            # May be merge on_client_connection_close and on_access_log???
            # probably by simply deprecating on_client_connection_close in future.
            self.route.on_client_connection_close()
            ctx = self.route.on_access_log(context)
            if ctx is None:
                log_handled = True
            else:
                context = ctx
        if not log_handled:
            self.access_log(context)

    def access_log(self, context: Dict[str, Any]) -> None:
        logger.info(DEFAULT_WEB_ACCESS_LOG_FORMAT.format_map(context))

    @property
    def _protocol(self) -> Tuple[bool, int]:
        do_ws_upgrade = self.request.is_websocket_upgrade
        return do_ws_upgrade, httpProtocolTypes.WEBSOCKET \
            if do_ws_upgrade \
            else httpProtocolTypes.HTTPS \
            if self.encryption_enabled() \
            else httpProtocolTypes.HTTP

    def _try_route(self, path: bytes) -> bool:
        do_ws_upgrade, protocol = self._protocol
        for route in self.routes[protocol]:
            if route.match(text_(path)):
                self.route = self.routes[protocol][route]
                assert self.route
                # Optionally, upgrade protocol
                if do_ws_upgrade and self.route.do_upgrade(self.request):
                    self.switch_to_websocket()
                    assert self.route
                    # Invoke plugin.on_websocket_open
                    self.route.on_websocket_open()
                else:
                    # Invoke plugin.handle_request
                    self.route.handle_request(self.request)
                    # if self.request.has_header(b'connection') and \
                    #         self.request.header(b'connection').lower() == b'close':
                    #     return True
                # Bailout on first match
                break
        return False

    def _try_static_or_404(self, path: bytes) -> None:
        path = text_(path).split('?', 1)[0]
        self.client.queue(
            HttpWebServerBasePlugin.serve_static_file(
                self.flags.static_server_dir + path,
                self.flags.min_compression_length,
            ),
        )

    def emit_request_complete(self) -> None:
        if not self.flags.enable_events:
            return
        assert self.request.port and self.event_queue
        self.event_queue.publish(
            request_id=self.uid,
            event_name=eventNames.REQUEST_COMPLETE,
            event_payload={
                'url': 'http://%s%s'
                % (
                    text_(self.request.header(b'host')),
                    text_(self.request.path),
                ),
                'method': text_(self.request.method),
                'headers': (
                    {}
                    if not self.request.headers
                    else {
                        text_(k): text_(v[1]) for k, v in self.request.headers.items()
                    }
                ),
                'body': (
                    text_(self.request.body, errors='ignore')
                    if self.request.method == httpMethods.POST
                    else None
                ),
            },
            publisher_id=self.__class__.__qualname__,
        )



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import os
import sys
import gzip
import json
import time
import pprint
import signal
import socket
import getpass
import logging
import argparse
import threading
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, List, Type, Tuple, Optional, cast

from .core.ssh import SshTunnelListener, SshHttpProtocolHandler
from .core.work import ThreadlessPool
from .core.event import EventManager
from .http.codes import httpStatusCodes
from .common.flag import FlagParser, flags
from .http.client import client
from .common.types import HostPort
from .common.utils import bytes_
from .core.work.fd import RemoteFdExecutor
from .http.methods import httpMethods
from .core.acceptor import AcceptorPool
from .core.listener import ListenerPool
from .core.ssh.base import BaseSshTunnelListener
from .common.plugins import Plugins
from .common.version import __version__
from .common.constants import (
    IS_WINDOWS, HTTPS_PROTO, DEFAULT_PLUGINS, DEFAULT_VERSION,
    DEFAULT_LOG_FILE, DEFAULT_PID_FILE, DEFAULT_LOG_LEVEL, DEFAULT_BASIC_AUTH,
    DEFAULT_LOG_FORMAT, DEFAULT_WORK_KLASS, DEFAULT_OPEN_FILE_LIMIT,
    DEFAULT_ENABLE_DASHBOARD, DEFAULT_ENABLE_SSH_TUNNEL,
    DEFAULT_SSH_LISTENER_KLASS,
)
from .core.event.metrics import MetricsEventSubscriber
from .http.parser.parser import HttpParser


if TYPE_CHECKING:   # pragma: no cover
    from .core.listener import TcpSocketListener
    from .core.ssh.base import BaseSshTunnelHandler


logger = logging.getLogger(__name__)


flags.add_argument(
    '--version',
    '-v',
    action='store_true',
    default=DEFAULT_VERSION,
    help='Prints proxy.py version.',
)

# TODO: Add --verbose option which also
# starts to log traffic flowing between
# clients and upstream servers.
flags.add_argument(
    '--log-level',
    type=str,
    default=DEFAULT_LOG_LEVEL,
    help='Valid options: DEBUG, INFO (default), WARNING, ERROR, CRITICAL. '
    'Both upper and lowercase values are allowed. '
    'You may also simply use the leading character e.g. --log-level d',
)

flags.add_argument(
    '--log-file',
    type=str,
    default=DEFAULT_LOG_FILE,
    help='Default: sys.stdout. Log file destination.',
)

flags.add_argument(
    '--log-format',
    type=str,
    default=DEFAULT_LOG_FORMAT,
    help='Log format for Python logger.',
)

flags.add_argument(
    '--open-file-limit',
    type=int,
    default=DEFAULT_OPEN_FILE_LIMIT,
    help='Default: 1024. Maximum number of files (TCP connections) '
    'that proxy.py can open concurrently.',
)

flags.add_argument(
    '--plugins',
    action='append',
    nargs='+',
    default=DEFAULT_PLUGINS,
    help='Comma separated plugins.  ' +
    'You may use --plugins flag multiple times.',
)

# TODO: Ideally all `--enable-*` flags must be at the top-level.
# --enable-dashboard is specially needed here because
# ProxyDashboard class is not imported anywhere.
#
# Due to which, if we move this flag definition within dashboard
# plugin, users will have to explicitly enable dashboard plugin
# to also use flags provided by it.
flags.add_argument(
    '--enable-dashboard',
    action='store_true',
    default=DEFAULT_ENABLE_DASHBOARD,
    help='Default: False.  Enables proxy.py dashboard.',
)

# NOTE: Same reason as mention above.
# Ideally this flag belongs to proxy auth plugin.
flags.add_argument(
    '--basic-auth',
    type=str,
    default=DEFAULT_BASIC_AUTH,
    help='Default: No authentication. Specify colon separated user:password '
    'to enable basic authentication.',
)

flags.add_argument(
    '--enable-ssh-tunnel',
    action='store_true',
    default=DEFAULT_ENABLE_SSH_TUNNEL,
    help='Default: False.  Enable SSH tunnel.',
)

flags.add_argument(
    '--work-klass',
    type=str,
    default=DEFAULT_WORK_KLASS,
    help='Default: ' + DEFAULT_WORK_KLASS +
    '.  Work klass to use for work execution.',
)

flags.add_argument(
    '--pid-file',
    type=str,
    default=DEFAULT_PID_FILE,
    help='Default: None. Save "parent" process ID to a file.',
)

flags.add_argument(
    '--openssl',
    type=str,
    default='openssl',
    help='Default: openssl. Path to openssl binary. ' +
    'By default, assumption is that openssl is in your PATH.',
)

flags.add_argument(
    '--data-dir',
    type=str,
    default=None,
    help='Default: ~/.proxypy. Path to proxypy data directory.',
)

flags.add_argument(
    '--ssh-listener-klass',
    type=str,
    default=DEFAULT_SSH_LISTENER_KLASS,
    help='Default: '
    + DEFAULT_SSH_LISTENER_KLASS
    + '.  An implementation of BaseSshTunnelListener',
)


class Proxy:
    """Proxy is a context manager to control proxy.py library core.

    By default, :class:`~proxy.core.pool.AcceptorPool` is started with
    :class:`~proxy.http.handler.HttpProtocolHandler` work class.
    By definition, it expects HTTP traffic to flow between clients and server.

    In ``--threadless`` mode and without ``--local-executor``,
    a :class:`~proxy.core.executors.ThreadlessPool` is also started.
    Executor pool receives newly accepted work by :class:`~proxy.core.acceptor.Acceptor`
    and creates an instance of work class for processing the received work.

    In ``--threadless`` mode and with ``--local-executor 0``,
    acceptors will start a companion thread to handle accepted
    client connections.

    Optionally, Proxy class also initializes the EventManager.
    A multi-process safe pubsub system which can be used to build various
    patterns for message sharing and/or signaling.
    """

    def __init__(self, input_args: Optional[List[str]] = None, **opts: Any) -> None:
        self.opts = opts
        self.flags = FlagParser.initialize(input_args, **opts)
        self.listeners: Optional[ListenerPool] = None
        self.executors: Optional[ThreadlessPool] = None
        self.acceptors: Optional[AcceptorPool] = None
        self.event_manager: Optional[EventManager] = None
        self.ssh_tunnel_listener: Optional[BaseSshTunnelListener] = None
        self.metrics_subscriber: Optional[MetricsEventSubscriber] = None

    def __enter__(self) -> 'Proxy':
        self.setup()
        return self

    def __exit__(self, *args: Any) -> None:
        self.shutdown()

    def setup(self) -> None:
        # TODO: Introduce cron feature
        # https://github.com/abhinavsingh/proxy.py/discussions/808
        #
        # TODO: Introduce ability to change flags dynamically
        # https://github.com/abhinavsingh/proxy.py/discussions/1020
        #
        # TODO: Python shell within running proxy.py environment
        # https://github.com/abhinavsingh/proxy.py/discussions/1021
        #
        self._write_pid_file()
        # We setup listeners first because of flags.port override
        # in case of ephemeral port being used
        self.listeners = ListenerPool(flags=self.flags)
        self.listeners.setup()
        # Override flags.port to match the actual port
        # we are listening upon.  This is necessary to preserve
        # the server port when `--port=0` is used.
        if not self.flags.unix_socket_path:
            self.flags.port = cast(
                'TcpSocketListener',
                self.listeners.pool[0],
            )._port
        # --ports flag can also use 0 as value for ephemeral port selection.
        # Here, we override flags.ports to reflect actual listening ports.
        ports = set()
        offset = 1 if self.flags.unix_socket_path else 0
        for index in range(offset, offset + len(self.flags.ports)):
            ports.add(
                cast(
                    'TcpSocketListener',
                    self.listeners.pool[index],
                )._port,
            )
        if self.flags.port in ports:
            ports.remove(self.flags.port)
        self.flags.ports = list(ports)
        # Write ports to port file
        self._write_port_file()
        # Setup EventManager
        if self.flags.enable_events:
            logger.info('Core Event enabled')
            self.event_manager = EventManager()
            self.event_manager.setup()
        event_queue = (
            self.event_manager.queue if self.event_manager is not None else None
        )
        # Setup remote executors only if
        # --local-executor mode isn't enabled.
        if self.remote_executors_enabled:
            self.executors = ThreadlessPool(
                flags=self.flags,
                event_queue=event_queue,
                executor_klass=RemoteFdExecutor,
            )
            self.executors.setup()
        # Setup acceptors
        self.acceptors = AcceptorPool(
            flags=self.flags,
            listeners=self.listeners,
            executor_queues=self.executors.work_queues if self.executors else [],
            executor_pids=self.executors.work_pids if self.executors else [],
            executor_locks=self.executors.work_locks if self.executors else [],
            event_queue=event_queue,
        )
        self.acceptors.setup()
        # Start SSH tunnel acceptor if enabled
        if self.flags.enable_ssh_tunnel:
            self.ssh_tunnel_listener = self._setup_tunnel(
                flags=self.flags,
                **self.opts,
            )
        if event_queue is not None and self.flags.enable_metrics:
            self.metrics_subscriber = MetricsEventSubscriber(
                event_queue,
                self.flags.metrics_lock,
            )
            self.metrics_subscriber.setup()
        # TODO: May be close listener fd as we don't need it now
        if threading.current_thread() == threading.main_thread():
            self._register_signals()

    @staticmethod
    def _setup_tunnel(
        flags: argparse.Namespace,
        ssh_handler_klass: Optional[Type['BaseSshTunnelHandler']] = None,
        ssh_listener_klass: Optional[Any] = None,
        **kwargs: Any,
    ) -> BaseSshTunnelListener:
        listener_klass = ssh_listener_klass or SshTunnelListener
        handler_klass = ssh_handler_klass or SshHttpProtocolHandler
        tunnel = cast(Type[BaseSshTunnelListener], listener_klass)(
            flags=flags,
            handler=handler_klass(flags=flags),
            **kwargs,
        )
        tunnel.setup()
        return tunnel

    def shutdown(self) -> None:
        if self.metrics_subscriber is not None:
            self.metrics_subscriber.shutdown()
        if self.flags.enable_ssh_tunnel:
            assert self.ssh_tunnel_listener is not None
            self.ssh_tunnel_listener.shutdown()
        assert self.acceptors
        self.acceptors.shutdown()
        if self.remote_executors_enabled:
            assert self.executors
            self.executors.shutdown()
        if self.flags.enable_events:
            assert self.event_manager is not None
            self.event_manager.shutdown()
        if self.listeners:
            self.listeners.shutdown()
            self._delete_port_file()
            self._delete_pid_file()

    @property
    def remote_executors_enabled(self) -> bool:
        return self.flags.threadless and \
            not self.flags.local_executor

    def _write_pid_file(self) -> None:
        if self.flags.pid_file:
            with open(self.flags.pid_file, 'wb') as pid_file:
                pid_file.write(bytes_(os.getpid()))

    def _delete_pid_file(self) -> None:
        if self.flags.pid_file \
                and os.path.exists(self.flags.pid_file):
            os.remove(self.flags.pid_file)

    def _write_port_file(self) -> None:
        if self.flags.port_file:
            with open(self.flags.port_file, 'wb') as port_file:
                if not self.flags.unix_socket_path:
                    port_file.write(bytes_(self.flags.port))
                    port_file.write(b'\n')
                for port in self.flags.ports:
                    port_file.write(bytes_(port))
                    port_file.write(b'\n')

    def _delete_port_file(self) -> None:
        if self.flags.port_file \
                and os.path.exists(self.flags.port_file):
            os.remove(self.flags.port_file)

    def _register_signals(self) -> None:
        # TODO: Define SIGUSR1, SIGUSR2
        signal.signal(signal.SIGINT, self._handle_exit_signal)
        signal.signal(signal.SIGTERM, self._handle_exit_signal)
        if not IS_WINDOWS:
            if hasattr(signal, 'SIGINFO'):
                signal.signal(      # pragma: no cover
                    signal.SIGINFO,       # pylint: disable=E1101
                    self._handle_siginfo,
                )
            signal.signal(signal.SIGHUP, self._handle_exit_signal)
            # TODO: SIGQUIT is ideally meant to terminate with core dumps
            signal.signal(signal.SIGQUIT, self._handle_exit_signal)

    @staticmethod
    def _handle_exit_signal(signum: int, _frame: Any) -> None:
        logger.debug('Received signal %d' % signum)
        sys.exit(0)

    def _handle_siginfo(self, _signum: int, _frame: Any) -> None:
        pprint.pprint(self.flags.__dict__)  # pragma: no cover


def sleep_loop(p: Optional[Proxy] = None) -> None:
    while True:
        try:
            time.sleep(1)
        except KeyboardInterrupt:
            break


def main(**opts: Any) -> None:
    with Proxy(sys.argv[1:], **opts) as p:
        sleep_loop(p)


def entry_point() -> None:
    main()


def grout() -> None:  # noqa: C901
    default_grout_tld = os.environ.get('JAXL_DEFAULT_GROUT_TLD', 'jaxl.io')

    def _clear_line() -> None:
        print('\r' + ' ' * 60, end='', flush=True)

    def _env(scheme: bytes, host: bytes, port: int) -> Optional[Dict[str, Any]]:
        try:
            response = client(
                scheme=scheme,
                host=host,
                port=port,
                path=b'/env/',
                method=httpMethods.BIND,
                body='v={0}&u={1}&h={2}'.format(
                    __version__,
                    os.environ.get('USER', getpass.getuser()),
                    socket.gethostname(),
                ).encode(),
            )
        except socket.gaierror:
            _clear_line()
            print(
                '\r\033[91mUnable to resolve\033[0m',
                end='',
                flush=True,
            )
            return None
        if response:
            if (
                response.code is not None
                and int(response.code) == httpStatusCodes.OK
                and response.body is not None
            ):
                return cast(
                    Dict[str, Any],
                    json.loads(
                        (
                            gzip.decompress(response.body).decode()
                            if response.has_header(b'content-encoding')
                            and response.header(b'content-encoding') == b'gzip'
                            else response.body.decode()
                        ),
                    ),
                )
            if response.code is None:
                _clear_line()
                print('\r\033[91mUnable to fetch\033[0m', end='', flush=True)
            else:
                _clear_line()
                print(
                    '\r\033[91mError code {0}\033[0m'.format(
                        response.code.decode(),
                    ),
                    end='',
                    flush=True,
                )
        else:
            _clear_line()
            print(
                '\r\033[91mUnable to connect\033[0m',
                end='',
                flush=True,
            )
        return None

    def _parse() -> Tuple[str, int]:
        """Here we deduce registry host/port based upon input parameters."""
        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument('route', nargs='?', default=None)
        parser.add_argument('name', nargs='?', default=None)
        parser.add_argument('--wildcard', action='store_true', help='Enable wildcard')
        args, _remaining_args = parser.parse_known_args()
        grout_tld = default_grout_tld
        if args.name is not None and '.' in args.name:
            grout_tld = args.name if args.wildcard else args.name.split('.', maxsplit=1)[1]
        grout_tld_parts = grout_tld.split(':')
        tld_host = grout_tld_parts[0]
        tld_port = 443
        if len(grout_tld_parts) > 1:
            tld_port = int(grout_tld_parts[1])
        return tld_host, tld_port

    tld_host, tld_port = _parse()
    env = None
    attempts = 0
    try:
        while True:
            env = _env(scheme=HTTPS_PROTO, host=tld_host.encode(), port=int(tld_port))
            attempts += 1
            if env is not None:
                print('\rStarting ...' + ' ' * 30 + '\r', end='', flush=True)
                break
            time.sleep(1)
            _clear_line()
            print(
                '\rWaiting for connection {0}'.format('.' * (attempts % 4)),
                end='',
                flush=True,
            )
            time.sleep(1)
    except KeyboardInterrupt:
        sys.exit(1)

    assert env is not None
    print('\r' + ' ' * 70 + '\r', end='', flush=True)
    Plugins.from_bytes(env['m'].encode(), name='client').grout(env=env['e'])  # type: ignore[attr-defined]


class GroutClientBasePlugin(ABC):
    """Base class for dynamic grout client rules.

    Implementation of this class must be stateless because a new instance is created
    for every route decision making.
    """

    @abstractmethod
    def resolve_route(
        self,
        route: str,
        request: HttpParser,
        origin: HostPort,
        server: HostPort,
    ) -> Tuple[Optional[str], HttpParser]:
        """Returns a valid grout route string.

        You MUST override this method.  This method returns 2-tuple where
        first value is the "route" and second the "request" object.

        For a simple pass through, simply return the "route" argument value itself.
        You can also return a dynamic value based upon "request" and "origin" information.
        E.g. sending to different upstream services based upon request Host header.
        Return None as "route" value to drop the request.

        You can also modify the original request object and return.  Common examples
        include strip-path scenario, where you would like to strip the path before
        sending the request to upstream.
        """
        raise NotImplementedError()



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       acceptor
"""
import argparse
from abc import ABC, abstractmethod
from uuid import uuid4
from typing import TYPE_CHECKING, Any, Dict, Generic, TypeVar, Optional

from ..event import EventQueue, eventNames
from ...common.types import Readables, Writables, SelectableEvents


if TYPE_CHECKING:   # pragma: no cover
    from ..connection import UpstreamConnectionPool

T = TypeVar('T')


class Work(ABC, Generic[T]):
    """Implement Work to hook into the event loop provided by Threadless process."""

    def __init__(
            self,
            work: T,
            flags: argparse.Namespace,
            event_queue: Optional[EventQueue] = None,
            uid: Optional[str] = None,
            upstream_conn_pool: Optional['UpstreamConnectionPool'] = None,
    ) -> None:
        # Work uuid
        self.uid: str = uid if uid is not None else uuid4().hex
        self.flags = flags
        # Eventing core queue
        self.event_queue = event_queue
        # Accept work
        self.work = work
        self.upstream_conn_pool = upstream_conn_pool

    @staticmethod
    @abstractmethod
    def create(*args: Any) -> T:
        """Implementations are responsible for creation of work objects
        from incoming args.  This helps keep work core agnostic to
        creation of externally defined work class objects."""
        raise NotImplementedError()

    async def get_events(self) -> SelectableEvents:
        """Return sockets and events (read or write) that we are interested in."""
        return {}   # pragma: no cover

    async def handle_events(
            self,
            _readables: Readables,
            _writables: Writables,
    ) -> bool:
        """Handle readable and writable sockets.

        Return True to shutdown work."""
        return False    # pragma: no cover

    def initialize(self) -> None:
        """Perform any resource initialization."""
        pass    # pragma: no cover

    def is_inactive(self) -> bool:
        """Return True if connection should be considered inactive."""
        return False    # pragma: no cover

    def shutdown(self) -> None:
        """Implementation must close any opened resources here
        and call super().shutdown()."""
        self.publish_event(
            event_name=eventNames.WORK_FINISHED,
            event_payload={},
            publisher_id=self.__class__.__qualname__,
        )

    def run(self) -> None:
        """run() method is not used by Threadless.  It's here for backward
        compatibility with threaded mode where work class is started as
        a separate thread.
        """
        pass    # pragma: no cover

    def publish_event(
            self,
            event_name: int,
            event_payload: Dict[str, Any],
            publisher_id: Optional[str] = None,
    ) -> None:
        """Convenience method provided to publish events into the global event queue."""
        if not self.flags.enable_events:
            return
        assert self.event_queue
        self.event_queue.publish(
            self.uid,
            event_name,
            event_payload,
            publisher_id,
        )



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
from ..http.proxy import HttpProxyBasePlugin
from ..http.parser import HttpParser


class TlsInterceptConditionallyPlugin(HttpProxyBasePlugin):
    """TLS intercept conditionally."""

    def do_intercept(self, request: HttpParser) -> bool:
        if request.host == b'httpbin.org':
            return False
        return super().do_intercept(request)



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import re
import ssl
import sys
import queue
import socket
import ipaddress
from typing import TYPE_CHECKING, Any, Dict, List, Tuple, Union, TypeVar


if TYPE_CHECKING:   # pragma: no cover
    DictQueueType = queue.Queue[Dict[str, Any]]
else:
    DictQueueType = queue.Queue


Selectable = int
Selectables = List[Selectable]
SelectableEvents = Dict[Selectable, int]    # Values are event masks
Readables = Selectables
Writables = Selectables
Descriptors = Tuple[Readables, Writables]
IpAddress = Union[ipaddress.IPv4Address, ipaddress.IPv6Address]
TcpOrTlsSocket = Union[ssl.SSLSocket, socket.socket]
HostPort = Tuple[str, int]

if sys.version_info.minor == 6:
    RePattern = TypeVar('RePattern', bound=Any)
elif sys.version_info.minor in (7, 8):
    RePattern = TypeVar('RePattern', bound=re.Pattern)  # type: ignore
else:
    RePattern = TypeVar('RePattern', bound=re.Pattern[Any])  # type: ignore



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       shortlink
"""
from typing import Optional

from ..http.proxy import HttpProxyBasePlugin
from ..http.parser import HttpParser
from ..http.responses import NOT_FOUND_RESPONSE_PKT, seeOthersResponse
from ..common.constants import DOT, SLASH


class ShortLinkPlugin(HttpProxyBasePlugin):
    """Add support for short links in your favorite browsers / applications.

    Enable ShortLinkPlugin and speed up your daily browsing experience.

    Example::
    * ``f/`` for ``facebook.com``
    * ``g/`` for ``google.com`
    * ``t/`` for ``twitter.com``
    * ``y/`` for ``youtube.com``
    * ``proxy/`` for ``py`` internal web servers.
    Customize map below for your taste and need.

    Paths are also preserved. E.g. ``t/imoracle`` will
    resolve to my Twitter profile for username ``imoracle``.
    """

    SHORT_LINKS = {
        b'a': b'amazon.com',
        b'i': b'instagram.com',
        b'l': b'linkedin.com',
        b'f': b'facebook.com',
        b'g': b'google.com',
        b't': b'twitter.com',
        b'w': b'web.whatsapp.com',
        b'y': b'youtube.com',
        b'proxy': b'localhost:8899',
    }

    def before_upstream_connection(
            self, request: HttpParser,
    ) -> Optional[HttpParser]:
        if request.host and request.host != b'localhost' and DOT not in request.host:
            # Avoid connecting to upstream
            return None
        return request

    def handle_client_request(
            self, request: HttpParser,
    ) -> Optional[HttpParser]:
        if request.host and request.host != b'localhost' and DOT not in request.host:
            if request.host in self.SHORT_LINKS:
                path = SLASH if not request.path else request.path
                self.client.queue(
                    seeOthersResponse(
                        b'http://' + self.SHORT_LINKS[request.host] + path,
                    ),
                )
            else:
                self.client.queue(NOT_FOUND_RESPONSE_PKT)
            return None
        return request



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
from abc import ABC, abstractmethod
from typing import Optional

from ....http.parser import HttpParser


class CacheStore(ABC):

    def __init__(self, uid: str) -> None:
        self.uid = uid

    @abstractmethod
    def open(self, request: HttpParser) -> None:
        pass

    @abstractmethod
    def cache_request(self, request: HttpParser) -> Optional[HttpParser]:
        return request

    @abstractmethod
    def cache_response_chunk(self, chunk: memoryview) -> memoryview:
        return chunk

    @abstractmethod
    def close(self) -> None:
        pass



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       ws
       onmessage
"""
import logging
from typing import List, Tuple

from ..http.parser import HttpParser
from ..http.server import HttpWebServerBasePlugin, httpProtocolTypes
from ..http.responses import okResponse
from ..http.websocket.frame import WebsocketFrame


logger = logging.getLogger(__name__)

HTTP_RESPONSE = okResponse(content=b'HTTP route response')
HTTPS_RESPONSE = okResponse(content=b'HTTPS route response')


class WebServerPlugin(HttpWebServerBasePlugin):
    """Demonstrates inbuilt web server routing using plugin."""

    def routes(self) -> List[Tuple[int, str]]:
        return [
            (httpProtocolTypes.HTTP, r'/http-route-example$'),
            (httpProtocolTypes.HTTPS, r'/https-route-example$'),
            (httpProtocolTypes.WEBSOCKET, r'/ws-route-example$'),
        ]

    def handle_request(self, request: HttpParser) -> None:
        if request.path == b'/http-route-example':
            self.client.queue(HTTP_RESPONSE)
        elif request.path == b'/https-route-example':
            self.client.queue(HTTPS_RESPONSE)

    def on_websocket_message(self, frame: WebsocketFrame) -> None:
        """Open chrome devtools and try using following commands:

        Example:

            ws = new WebSocket("ws://localhost:8899/ws-route-example")
            ws.onmessage = function(m) { console.log(m); }
            ws.send('hello')

        """
        self.client.queue(memoryview(WebsocketFrame.text(frame.data or b'')))
